#
# Copyright (c) 2024 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys
import shutil
import subprocess

# import yaml
import argparse
from typing import Optional, Union
from pathlib import Path

from isaac_toolkit.session import Session
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()

DEFAULT_DOCKER_IMAGE = "isaac-quickstart-spike:latest"


def retarget_spike_iss(
    sess: Session,
    workdir: Optional[Union[str, Path]] = None,
    mount_dir: Optional[Union[str, Path]] = None,
    docker_image: Optional[str] = None,
    spike_core: Optional[str] = None,
    label: Optional[str] = None,
    force: bool = False,
    verbose: bool = False,
    cleanup: bool = False,
):
    assert workdir is not None
    if not isinstance(workdir, Path):
        workdir = Path(workdir)
    assert workdir.is_dir()
    if spike_core is None:
        spike_core = "XIsaacCore"
    if label is None:
        label = "default"
    use_docker = docker_image is not None
    subdir = "docker" if use_docker else "local"
    base_dir = workdir / subdir
    output_dir = (base_dir / "spike") if label == "" else (base_dir / f"spike_{label}")
    if output_dir.is_dir():
        assert force, f"Directory already exists: {output_dir}. Use --force or different --label."
        logger.info("Cleaning up old output dir: %s (--force)", output_dir)
        shutil.rmtree(output_dir)
    output_dir.mkdir(parents=True)
    # gen_dir = workdir / "gen" / label
    gen_dir = (workdir / "gen") if label == "" else (workdir / f"gen_{label}")
    top_file = gen_dir / f"{spike_core}.core_desc"
    kwargs = {}
    # print("verbose", verbose)
    if not verbose:
        kwargs.setdefault("stdout", subprocess.PIPE)
        kwargs.setdefault("stderr", subprocess.PIPE)
        # kwargs.setdefault("text", True)
    if use_docker:
        command = "docker run -it --rm"
        if mount_dir is not None:
            command += f" -v {mount_dir}:{mount_dir}"
        command += f" -e CLEANUP={int(cleanup)}"
        command += f" {docker_image}"
        command += f" {output_dir}"
        command += f" {top_file}"
        # print("$$$", command)

        # input("!")
        try:
            # print("command", command)
            # input(">")
            subprocess.run(command, check=True, shell=True, **kwargs)
        except subprocess.CalledProcessError as e:
            print(f"[ERROR] Command failed with return code {e.returncode}")
            if e.stdout:
                print("--- STDOUT ---")
                print(e.stdout.decode())
            if e.stderr:
                print("--- STDERR ---")
                print(e.stderr.decode())
            raise  # Re-raise if you want the caller to handle it too
    else:

        temp_dir = base_dir / "temp"
        spike_home = temp_dir / "spike"
        env = os.environ.copy()
        env["SPIKE_HOME"] = spike_home
        env["CLEANUP"] = str(int(cleanup))
        # ccache already exported implicitly?
        # TODO: explicit ccache?
        # env["CCACHE"]
        # env["CCACHE_DIR"]
        spike_script = os.environ.get("SPIKE_SCRIPT_LOCAL", None)
        assert spike_script is not None, "SPIKE_SCRIPT_LOCAL undefined"
        assert Path(spike_script).is_file(), f"Not found: {spike_script}"
        # TODO: ship with isaac?
        spike_script_args = [output_dir, top_file]
        try:
            subprocess.run([spike_script, *spike_script_args], check=True, **kwargs, env=env)
        except subprocess.CalledProcessError as e:
            print(f"[ERROR] Command failed with return code {e.returncode}")
            if e.stdout:
                print("--- STDOUT ---")
                print(e.stdout.decode())
            if e.stderr:
                print("--- STDERR ---")
                print(e.stderr.decode())
            raise  # Re-raise if you want the caller to handle it too


def handle(args):
    # assert args.session is not None
    sess = None
    if args.session is not None:
        session_dir = Path(args.session)
        assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
        sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    retarget_spike_iss(
        sess,
        force=args.force,
        workdir=args.workdir,
        docker_image=args.docker,
        verbose=args.verbose,
    )
    if sess is not None:
        sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    # parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--session", "--sess", "-s", type=str, required=False)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--docker", type=str, default=None, const=DEFAULT_DOCKER_IMAGE, nargs="?")
    parser.add_argument("--workdir", type=str, default=None)
    parser.add_argument("--verbose", action="store_true")
    # label: Optional[str] = None,
    # etiss_core: Optional[str] = None,

    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import TableArtifact, filter_artifacts
from .check_ise_potential import get_unsupported_opcodes, get_ise_potential_df
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def check_ise_potential_per_llvm_bb(
    sess: Session,
    min_supported: float = 0.15,
    allow_mem: bool = False,
    allow_loads: bool = False,
    allow_stores: bool = False,
    allow_branches: bool = False,
    allow_compressed: bool = True,
    allow_custom: bool = True,
    allow_fp: bool = False,
    allow_rvv: bool = False,
    allow_system: bool = False,
    force: bool = False,
):
    logger.info("Checking ISE potential per LLVM BB...")
    artifacts = sess.artifacts
    opcodes_hist_artifacts = filter_artifacts(artifacts, lambda x: x.name == "opcodes_per_llvm_bb_hist")
    assert len(opcodes_hist_artifacts) == 1
    opcodes_hist_artifact = opcodes_hist_artifacts[0]

    opcodes_per_llvm_bb_hist_df = opcodes_hist_artifact.df
    unsupported_opcodes = get_unsupported_opcodes(
        allow_mem=allow_mem,
        allow_loads=allow_loads,
        allow_stores=allow_stores,
        allow_branches=allow_branches,
        allow_compressed=allow_compressed,
        allow_custom=allow_custom,
        allow_fp=allow_fp,
        allow_rvv=allow_rvv,
        allow_system=allow_system,
    )

    dfs = []
    for group, opcodes_hist_df in opcodes_per_llvm_bb_hist_df.groupby(["func_name", "bb_name"]):
        func_name, bb_name = group
        # print("func_name", func_name)
        # print("bb_name", bb_name)
        # print("opcodes_hist_df")
        # print(opcodes_hist_df)
        # force_rvv = True
        force_rvv = False
        forced_opcodes = []
        if force_rvv:
            forced_opcodes = ["OP-V"]
        if forced_opcodes:
            forced_opcodes_hist_df = opcodes_hist_df[opcodes_hist_df["opcode"].isin(forced_opcodes)]
            if len(forced_opcodes_hist_df) == 0:
                # TODO: logging
                continue

        ise_potential_df = get_ise_potential_df(opcodes_hist_df, unsupported_opcodes, min_supported)
        ise_potential_df.insert(0, "func_name", func_name)
        ise_potential_df.insert(1, "bb_name", bb_name)
        # print("ise_potential_df")
        # print(ise_potential_df)
        dfs.append(ise_potential_df)
    ise_potential_per_llvm_bb_df = pd.concat(dfs)
    # print(ise_potential_per_llvm_bb_df)
    # input(">")

    attrs = {
        "kind": "table",
        "by": __name__,
    }

    artifact = TableArtifact("ise_potential_per_llvm_bb", ise_potential_per_llvm_bb_df, attrs=attrs)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    check_ise_potential_per_llvm_bb(
        sess,
        allow_mem=args.allow_mem,
        allow_loads=args.allow_loads,
        allow_stores=args.allow_stores,
        allow_branches=args.allow_branches,
        allow_compressed=args.allow_compressed,
        allow_custom=args.allow_custom,
        allow_fp=args.allow_fp,
        allow_rvv=args.allow_rvv,
        allow_system=args.allow_system,
        force=args.force,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: default to allowed?
    parser.add_argument("--min-supported", type=float, default=0.15)
    parser.add_argument("--allow-mem", action="store_true")
    parser.add_argument("--allow-loads", action="store_true")
    parser.add_argument("--allow-stores", action="store_true")
    parser.add_argument("--allow-branches", action="store_true")
    parser.add_argument("--allow-compressed", action="store_true")
    parser.add_argument("--allow-custom", action="store_true")
    parser.add_argument("--allow-fp", action="store_true")
    parser.add_argument("--allow-rvv", action="store_true")
    parser.add_argument("--allow-system", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import logging
import argparse
from pathlib import Path

import pandas as pd
from elftools.elf.elffile import ELFFile
from elftools.elf.sections import SymbolTableSection

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts


logger = logging.getLogger("symbol_table")


# TODO: move dwarf.py in elf subdir


def parse_elf(elf_path):
    with open(elf_path, "rb") as f:
        elffile = ELFFile(f)

        data = []

        section = elffile.get_section_by_name(".symtab")

        assert section, "Symbol Table not found!"
        assert isinstance(section, SymbolTableSection)
        for sym in section.iter_symbols():
            ty = sym.entry["st_info"]["type"]
            # if ty != "STT_FUNC":
            #     continue
            name = sym.name
            sz = sym.entry["st_size"]
            val = sym.entry["st_value"]
            print("ty", ty)
            print("name", name)
            print("sz", sz)
            print("val", val)
            new = {"name": name, "type": ty, "size": sz, "value": val}
            data.append(new)
        symbol_table_df = pd.DataFrame(data)
    return symbol_table_df


def analyze_dwarf(sess: Session, force: bool = False):
    artifacts = sess.artifacts
    elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    assert len(elf_artifacts) == 1
    elf_artifact = elf_artifacts[0]

    symbol_table_df = parse_elf(elf_artifact.path)

    attrs = {
        "elf_file": elf_artifact.name,
        "kind": "mapping",
        "by": "isaac_toolkit.analysis.static.elf.symbol_table",
    }

    symbol_table_artifact = TableArtifact("symbol_table", symbol_table_df, attrs=attrs)
    sess.add_artifact(symbol_table_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    analyze_dwarf(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from isaac_toolkit.session.artifact import (
    ArtifactFlag,
    PythonArtifact,
    filter_artifacts,
)


FOO_ARTIFACT = PythonArtifact("foo", data=[])
BAR_ARTIFACT = PythonArtifact("bar", data=[], attrs={"key": "val"})
BAZ_ARTIFACT = PythonArtifact("baz", data=[], flags=ArtifactFlag.TABLE)

ALL_ARTIFACTS = [FOO_ARTIFACT, BAR_ARTIFACT, BAZ_ARTIFACT]


@pytest.mark.parametrize(
    "artifacts,func,expected",
    [
        (ALL_ARTIFACTS, lambda _: True, ALL_ARTIFACTS),
        (ALL_ARTIFACTS, lambda _: False, []),
        (ALL_ARTIFACTS, lambda x: x.name == "foo", [FOO_ARTIFACT]),
        (ALL_ARTIFACTS, lambda x: x.attrs.get("key") == "val", [BAR_ARTIFACT]),
        (ALL_ARTIFACTS, lambda x: x.flags & ArtifactFlag.PYTHON, ALL_ARTIFACTS),
        (ALL_ARTIFACTS, lambda x: x.flags & ArtifactFlag.TABLE, [BAZ_ARTIFACT]),
        ([], lambda x: x.name in ["foo", "bar", "baz"], []),
    ],
)
def test_artifact_filter_artifacts(artifacts, func, expected):
    assert filter_artifacts(artifacts, func) == expected



import sys
import argparse
from pathlib import Path

import pandas as pd


def analyze_compare_perf(compare_perf_csv, uarchs_csv, output=None, print_df=False):

    compare_perf_csv = Path(compare_perf_csv)
    assert compare_perf_csv.is_file()
    compare_perf_df = pd.read_csv(compare_perf_csv)
    if "Unnamed: 0" in compare_perf_df.columns:
        compare_perf_df.drop(columns=["Unnamed: 0"], inplace=True)
    print(compare_perf_df, compare_perf_df.columns, len(compare_perf_df))

    uarchs_csv = Path(uarchs_csv)
    assert uarchs_csv.is_file()
    uarchs_df = pd.read_csv(uarchs_csv)
    # uarchs_df.rename(index={0: "idx"}, inplace=True)
    if "Unnamed: 0" in uarchs_df.columns:
        uarchs_df.drop(columns=["Unnamed: 0"], inplace=True)
    print(uarchs_df, uarchs_df.columns, len(uarchs_df))

    uarchs_bench_df = compare_perf_df[["uArch", "Run Cycles", "Run Cycles (rel.)", "Run CPI"]]
    uarchs_bench_df = uarchs_bench_df.iloc[1:]
    sorted_uarchs_bench_df = uarchs_bench_df.sort_values("Run CPI")
    print("uarchs_bench_df")
    print(uarchs_bench_df)
    print("sorted_uarchs_bench_df")
    print(sorted_uarchs_bench_df)
    sorted_uarchs = list(sorted_uarchs_bench_df["uArch"].values)
    sorted_cpis = list(sorted(set(list(sorted_uarchs_bench_df["Run CPI"].values))))
    # TODO: same rank if same score?
    # TODO: replace CPI with Score
    print("sorted_uarchs", sorted_uarchs)
    print("sorted_cpis", sorted_cpis)
    merged_df = pd.merge(uarchs_bench_df, uarchs_df, left_on="uArch", right_on="uarch", how="inner", suffixes=("", ""))
    merged_df.drop(columns=["uarch"], inplace=True)
    # merged_df["Rank"] = merged_df["uArch"].apply(lambda x: sorted_uarchs.index(x) + 1)
    merged_df["Rank"] = merged_df["Run CPI"].apply(lambda x: sorted_cpis.index(x) + 1)
    merged_df = merged_df.sort_values("Rank")
    # merged_df.drop(columns=["Run Cycles", "Run Cycles (rel.)", "Run CPI", "uarch_lower", "variant"], inplace=True)
    merged_df.drop(columns=["Run Cycles", "Run Cycles (rel.)", "uarch_lower", "variant"], inplace=True)
    # TODO: attach costs

    df = merged_df

    assert print_df or output is not None

    if print_df:
        with pd.option_context(
            "display.max_columns", None, "display.width", 1000, "display.max_colwidth", 400, "display.min_rows", 100
        ):
            print(df)

    if output is not None:
        # TODO: assert csv suffix?
        df.to_csv(output)


def handle(args):
    analyze_compare_perf(
        args.report,
        args.uarchs_csv,
        output=args.output,
        print_df=args.print_df,
    )


def get_parser():
    parser = argparse.ArgumentParser(description="Collect metrics from MLonMCU Report")
    parser.add_argument("report", help="Compare Perf CSV file")
    parser.add_argument("--uarchs-csv", default=None, required=True, help="uArchs CSV file")
    parser.add_argument("-o", "--output", default=None, help="Output file")
    parser.add_argument("--print-df", action="store_true", help="Print DataFrame")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from typing import Optional
from pathlib import Path

from isaac_toolkit.logging import get_logger, set_log_level
from . import Session

logger = get_logger()


def create(session_dir: Path, force: bool = False, interactive: bool = False, default_log_level: Optional[str] = None):
    del force
    if session_dir.is_dir():
        logger.info("Re-initializing existing session: %s", session_dir)
        if interactive:
            raise NotImplementedError("Interactive mode")
        sess = Session.from_dir(session_dir)
    else:
        logger.info("Initializing new session: %s", session_dir)
        sess = Session.create(session_dir)
    if default_log_level is not None:
        sess.config.logging.console.level = default_log_level
        sess.config.logging.file.level = default_log_level
        set_log_level(console_level=default_log_level, file_level=default_log_level)
        sess.save()
    return sess


def handle_create(args):
    assert args.session is not None
    session_dir = Path(args.session)
    force = args.force
    interactive = not force
    create(session_dir, force=force, interactive=interactive, default_log_level=args.log)


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle_create(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import argparse

import yaml
import pandas as pd


def names_helper(index, output=None):
    with open(index, "r") as f:
        combined_index_data = yaml.safe_load(f)
        # TODO: index and cdsl should use the same instruction names?
        names = [
            candidate["properties"].get("InstrName", f"CUSTOM{i}")
            for i, candidate in enumerate(combined_index_data["candidates"])
        ]
        print("names", names)
        # num_candidates = len(names)
        names_df = pd.DataFrame({"instr": names})
        names_df["instr_lower"] = names_df["instr"].apply(lambda x: x.lower())
        names_df["idx"] = names_df["instr"].apply(lambda x: names.index(x))

    if output is None:
        print(names_df)
    else:
        names_df.to_csv(output, index=False)


def handle(args):
    names_helper(args.index, output=args.output)


def get_parser():
    parser = argparse.ArgumentParser(description="Extract instr names from index YAML and write to CSV")
    parser.add_argument("index", help="Index yaml file")
    parser.add_argument("-o", "--output", default=None, help="Output CSV file")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd
from elftools.elf.elffile import ELFFile

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def parse_elf(elf_path):
    with open(elf_path, "rb") as f:
        elffile = ELFFile(f)
        ###
        from elftools.elf.sections import SymbolTableSection

        # print('  %s sections' % elffile.num_sections())
        section = elffile.get_section_by_name(".symtab")

        assert section, "Symbol Table not found!"
        # print('  Section name: %s, type: %s' %(section.name, section['sh_type']))
        if isinstance(section, SymbolTableSection):
            total_footprint = 0
            func_footprint = {}
            for i, sym in enumerate(section.iter_symbols()):
                # print("i", s]ym.entry)
                ty = sym.entry["st_info"]["type"]
                if ty != "STT_FUNC":
                    continue
                func = sym.name
                sz = sym.entry["st_size"]
                # print("ty", ty)
                # print("sz", sz)
                func_footprint[func] = sz
                total_footprint += sz
            # print("total_footprint", total_footprint)
            # print("func_footprint", func_footprint)
            footprint_df = pd.DataFrame(func_footprint.items(), columns=["func", "bytes"])
            footprint_df.sort_values("bytes", inplace=True, ascending=False)
            footprint_df["rel_bytes"] = footprint_df["bytes"] / total_footprint
            # print("footprint_df", footprint_df)
            # print("  The name of the last symbol in the section is: %s" % (section.get_symbol(num_symbols - 1).name))
        # input("123")
        ###
        return footprint_df


def analyze_mem_footprint(sess: Session, force: bool = False):
    logger.info("Analyzing memory footprint...")
    artifacts = sess.artifacts
    # print("artifacts", artifacts)
    elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    # print("elf_artifacts", elf_artifacts)
    assert len(elf_artifacts) == 1
    elf_artifact = elf_artifacts[0]

    footprint_df = parse_elf(elf_artifact.path)

    attrs = {
        "elf_file": elf_artifact.name,
        "kind": "mem_footprint",
        "by": "isaac_toolkit.analysis.static.mem_footprint",
    }

    artifact = TableArtifact("mem_footprint", footprint_df, attrs=attrs)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    analyze_mem_footprint(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import argparse
from pathlib import Path
from collections import defaultdict
from typing import Dict, Optional

import matplotlib.pyplot as plt
import pandas as pd
import yaml


def agg_util_scores(util_score_csv, output=None):

    dfs = []

    for score_csv in util_score_csv:
        df_ = pd.read_csv(score_csv)
        dfs.append(df_)

    df = pd.concat(dfs)
    # print("df", df)

    df = df.groupby("instr").sum().reset_index()
    # df = df.sort_values("util_score", ascending=False)
    df = df.sort_values("dynamic_util_score", ascending=False)
    # print("df", df)

    if output is None:
        print(df)
    else:
        df.to_csv(output, index=False)


def handle(args):
    agg_util_scores(args.util_score_csv, output=args.output)


def get_parser():
    parser = argparse.ArgumentParser(description="TODO")
    parser.add_argument("util_score_csv", nargs="+", help="CSV files")
    parser.add_argument("-o", "--output", default=None, help="Output CSV file")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys

from .c import get_parser, handle


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args, fmt="cpp")


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import DisassArtifact, TableArtifact
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def load_disass(sess: Session, input_file: Path, force: bool = False):
    logger.info("Loading Disassembly...")
    assert input_file.is_file(), f"File not found: {input_file}"
    name = input_file.name
    attrs = {
        "target": "riscv",  # TODO: 32/64?
        "kind": "disass",
        "elf": None,  # TODO
        "by": "isaac_toolkit.frontend.disass.objdump",
    }
    artifact = DisassArtifact(name, input_file, attrs=attrs)
    sess.add_artifact(artifact, override=force)

    with open(input_file, "r") as f:
        lines = f.readlines()
    data = []
    for line in lines:
        line = line.strip()
        if ":" not in line:
            continue
        addr, rest = line.split(":")
        addr = addr.strip()
        # print("rest", rest)
        splitted = rest.split("\t")
        splitted = list(filter(None, splitted))
        if len(splitted) < 2 or len(splitted) > 3:
            continue
        # print("splitted", splitted)
        addr = int(addr, 16)
        rest = rest.strip()
        # print("addr", addr)
        word = splitted[0]
        word = word.strip()
        word = word.replace(" ", "")
        word = int(word, 16)
        # print("word", word)
        insn = splitted[1]
        insn = insn.strip()
        # print("insn", insn)
        if len(splitted) == 3:
            args = splitted[2]
            args = args.strip()
        else:
            args = ""
        # print("args", args)
        # input(">")
        new = {"pc": addr, "bytecode": word, "instr": insn, "args": args}
        data.append(new)
    disass_df = pd.DataFrame(data)
    # name_ = f"{name}_table"
    disass_artifact = TableArtifact("disass", disass_df, attrs=attrs)
    sess.add_artifact(disass_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    input_file = Path(args.file)
    load_disass(sess, input_file, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# import time
import io
import sys
import argparse
import multiprocessing
from typing import List, Union, Optional
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

import pandas as pd
from tqdm import tqdm
from elftools.elf.elffile import ELFFile
from elftools.elf.constants import SH_FLAGS

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import InstrTraceArtifact

from .helper import process_df

# from pandarallel import pandarallel

# pandarallel.initialize()

# TODO: logger


class ELFInstructionFetcher:
    def __init__(self, elf_path):
        self.sections = []
        with open(elf_path, "rb") as f:
            elf = ELFFile(f)
            for section in elf.iter_sections():
                flags = section["sh_flags"]
                if flags & SH_FLAGS.SHF_EXECINSTR:
                    base = section["sh_addr"]
                    size = section["sh_size"]
                    data = section.data()
                    self.sections.append((base, base + size, data))

    def read_word_at_pc(self, pc, size=4):
        """Fetch `size` bytes at `pc` from any executable section."""
        for start, end, data in self.sections:
            if start <= pc < end:
                offset = pc - start
                if offset + size > len(data):
                    return None  # avoid out-of-bounds read
                return int.from_bytes(data[offset : offset + size], "little")
        return None  # pc not found


def disassemble_word(md, pc, word, size=4, endian="little", operands: bool = False):
    # print("disassemble_word", pc, word, size)
    try:
        code = word.to_bytes(size, endian)
        # print("code", code)
        insn = next(md.disasm(code, pc))
        # print("insn", insn)
        # input("%1")
        if operands:
            return insn.mnemonic, insn.op_str
        return insn.mnemonic
    except Exception:
        # print(e)
        # input("%2")
        if operands:
            return "unknown", ""
        return "unknown"


def parse_and_process(chunk_bytes):
    df = pd.read_csv(io.BytesIO(chunk_bytes), header=None, sep=",")
    return process_df(df)


def chunk_iter(path, chunk_size=2**22):
    with open(path, "rb") as f:
        buf = b""
        while True:
            data = f.read(chunk_size)
            if not data:
                break
            buf += data
            last_nl = buf.rfind(b"\n")
            if last_nl == -1:
                continue
            yield buf[: last_nl + 1]
            buf = buf[last_nl + 1 :]
        if buf:
            yield buf


def check_compressed(p):
    with open(p, "rb") as test_f:
        head = test_f.read(4)

    is_gzip = head.startswith(b"\x1f\x8b")
    is_lz4 = head == b"\x04\x22\x4d\x18"

    if is_gzip:
        return "gzip"
    if is_lz4:
        return "lz4"
    return "none"


def chunk_iter_compressed(path, chunk_size=2**22):
    print("chunk_iter_compressed")
    fmt = check_compressed(path)
    if fmt == "gzip":
        import gzip

        f = gzip.open(path, "rb")
    elif fmt == "lz4":
        import lz4.frame

        f = lz4.frame.open(path, "rb")
    else:
        f = open(path, "rb")

    with f:
        buf = b""
        while True:
            data = f.read(chunk_size)
            if not data:
                break
            buf += data
            last_nl = buf.rfind(b"\n")
            if last_nl == -1:
                continue
            yield buf[: last_nl + 1]
            buf = buf[last_nl + 1 :]
        if buf:
            yield buf


def load_instr_trace(
    sess: Session,
    input_files: Union[Path, List[Path]],
    force: bool = False,
    operands: bool = False,
    num_workers: Optional[int] = None,
    executor: str = "process_pool",
    chunk_size: int = 2**22,
):
    if num_workers is None:
        num_workers = multiprocessing.cpu_count()
    if operands:
        raise NotImplementedError("operands not included in pctrace")
    if not isinstance(input_files, list):
        input_files = [input_files]
    assert len(input_files) > 0
    name = input_files[0].name

    # elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    # print("elf_artifacts", elf_artifacts)
    # assert len(elf_artifacts) == 1, "TGC pctrace needs ELF artifact to lookup instructions"
    # elf_artifact = elf_artifacts[0]

    # sort input files by name
    sorted_files = sorted(input_files, key=lambda x: x.name)
    # df = pd.read_csv(input_file, sep=":", names=["pc", "rest"])
    dfs = []
    # with open(elf_artifact.path, "rb") as f:
    #     elf = ELFFile(f)
    # xlen = elf.elfclass
    # mode = CS_MODE_RISCV32 if xlen == 32 else CS_MODE_RISCV64

    # md = Cs(CS_ARCH_RISCV, mode | CS_MODE_RISCVC)
    # md.detail = False

    # fetcher = ELFInstructionFetcher(elf_artifact.path)
    # pc2bytecode = {}

    # def lookup_bytecode(pc, size):
    #     pc = int(pc)
    #     bytecode = pc2bytecode.get(pc)
    #     if bytecode is None:  # not in cache
    #         bytecode = fetcher.read_word_at_pc(pc, size=size)
    #         assert bytecode is not None
    #         pc2bytecode[pc] = bytecode
    #     return bytecode

    # bytecode2name = {}

    # def lookup_name(bytecode, pc=None, size=None):
    #     bytecode = int(bytecode)
    #     name = bytecode2name.get(bytecode)
    #     if name is None:  # not in cache
    #         assert pc is not None
    #         assert size is not None
    #         name = disassemble_word(pc, bytecode, size=size)
    #         assert name is not None
    #         bytecode2name[bytecode] = name
    #     # print("name", name)
    #     # input("?")
    #     return name

    # print("A", time.time())
    # t0 = time.time()
    # executor =
    executor_map = {
        "thread_pool": ThreadPoolExecutor,
        "process_pool": ProcessPoolExecutor,
    }
    executor_cls = executor_map.get(executor)
    assert executor_cls is not None, f"Unsupported Executor: {executor}"
    with executor_cls(max_workers=num_workers) as executor:  # tune workers
        # futures = []
        for input_file in sorted_files:
            assert input_file.is_file()
            # print("file", input_file)
            # with pd.read_csv(input_file, sep=",", chunksize=2**20, header=None) as reader:
            # with smart_open(input_file, "rb") as f:
            if True:
                # with pd.read_csv(input_file, sep=",", chunksize=2**22, header=None) as reader:
                # with pd.read_csv(f, sep=",", chunksize=2**22, header=None) as reader:
                # dfs_ = list(executor.map(parse_and_process, chunk_iter(f)))
                dfs_ = list(
                    tqdm(
                        executor.map(parse_and_process, chunk_iter_compressed(input_file, chunk_size=chunk_size)),
                        disable=False,
                    )
                )
                dfs += dfs_
                # with pd.read_csv(f, sep=",", chunksize=2**22, header=None) as reader:
                #     # for df in tqdm(reader, disable=False):
                #     for df in reader:
                #         # print("a")
                #         # future = executor.submit(process_df, df, fetcher, md)
                #         # dfs.append(process_df(df))
                #         future = executor.submit(process_df, df)
                #         futures.append(future)
        # for future in tqdm(as_completed(futures), total=len(futures)):
        #     # print("b")
        #     dfs.append(future.result())
    df = pd.concat(dfs, axis=0)
    # df["instr"] = df["instr"].astype("category")
    df["size"] = df["size"].astype("category")
    df["is_branch"] = df["is_branch"].astype("category")
    df["pc"] = pd.to_numeric(df["pc"], downcast="unsigned")
    # df["bytecode"] = pd.to_numeric(df["bytecode"], downcast="unsigned")
    df.reset_index(drop=True, inplace=True)
    # print("B", time.time() - t0)
    # print("df", df.head(), df.columns, df.dtypes, df.memory_usage())
    # input(">")

    attrs = {
        "simulator": "tgc",
        "cpu_arch": "unknown",
        "by": "isaac_toolkit.frontend.instr_trace.tgc",
    }
    artifact = InstrTraceArtifact(name, df, attrs=attrs)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    input_files = list(map(Path, args.files))
    load_instr_trace(
        sess,
        input_files,
        force=args.force,
        operands=args.operands,
        num_workers=args.num_workers,
        executor=args.executor,
        chunk_size=args.chunk_size,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("files", nargs="+")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--operands", action="store_true")
    parser.add_argument("--num-workers", type=int, default=multiprocessing.cpu_count())
    parser.add_argument("--chunk-size", type=int, default=2**22)
    parser.add_argument("--executor", choices=["process_pool", "thread_pool"], type=str, default="process_pool")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import argparse

import pandas as pd
import yaml


def assign_names(index, output=None, inplace=False, prefix="CUSTOM", csv=None, pkl=None):

    with open(index, "r") as f:
        combined_index_data = yaml.safe_load(f)
    candidates = combined_index_data["candidates"]
    # num_candidates = len(candidates)
    # print("candidates", candidates)

    names_data = []

    for i, candidate in enumerate(candidates):
        # print("i", i)
        properties = candidate.get("properties", {})
        # print("properties", properties)
        new_name = f"{prefix}{i}"
        properties["InstrName"] = new_name
        candidate["properties"] = properties
        num_fused_instrs = properties["#Instrs"]
        new_data = {"instr": new_name, "instr_lower": new_name.lower(), "idx": i, "num_fused_instrs": num_fused_instrs}
        names_data.append(new_data)

    names_df = pd.DataFrame(names_data)
    # print(names_df)

    if csv is not None:
        names_df.to_csv(csv, index=False)

    if pkl is not None:
        names_df.to_pickle(pkl)

    if inplace:
        assert output is None
        out_file = index
    else:
        assert output is not None
        out_file = output

    with open(out_file, "w") as f:
        yaml.dump(combined_index_data, f)


def handle(args):
    assign_names(args.index, output=args.output, inplace=args.inplace, prefix=args.prefix, csv=args.csv, pkl=args.pkl)


def get_parser():
    parser = argparse.ArgumentParser(description="TODO")
    parser.add_argument("index", help="Index yaml file")
    parser.add_argument("-o", "--output", default=None, help="Output yaml file")
    parser.add_argument("--inplace", action="store_true", help="TODO")
    parser.add_argument("--prefix", default="CUSTOM", help="TODO")
    parser.add_argument("--csv", default=None, help="TODO")
    parser.add_argument("--pkl", default=None, help="TODO")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys
import yaml
import argparse
import subprocess
from typing import Optional, Union
from pathlib import Path
from collections import defaultdict

import pandas as pd
from neo4j import GraphDatabase, Query
import networkx as nx
import networkx.algorithms.isomorphism as iso


from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts, TableArtifact
from isaac_toolkit.utils.graph_utils import memgraph_to_nx
from isaac_toolkit.algorithm.ise.identification.maxmiso import maxmiso_algo
from isaac_toolkit.logging import get_logger, set_log_level
from isaac_toolkit.utils.assign_names import assign_names
from isaac_toolkit.utils.combine_pdfs import combine_pdfs

logger = get_logger()


def get_unique_maxmisos(maxmisos):
    isos_map = defaultdict(list)
    covered = set()
    for i, m1 in enumerate(maxmisos):
        if i in covered:
            continue
        for j, m2 in enumerate(maxmisos):

            if j <= i:
                continue

            def node_match(*args):
                ret = iso.categorical_node_match("label", None)(*args)
                return ret

            def edge_match(*args):
                return iso.categorical_edge_match("label", None)(*args)

            is_iso = nx.is_isomorphic(m1, m2, node_match=node_match, edge_match=edge_match)
            if is_iso:
                isos_map[i].append(j)
                covered.add(j)

    logger.debug("isos_map", isos_map, len(isos_map))
    isos_size = {k: len(v) for k, v in isos_map.items()}
    logger.debug("isos_size", isos_size)
    logger.debug("covered", covered, len(covered))
    non_isos = [i for i in range(len(maxmisos)) if i not in covered]
    logger.debug("non_isos", non_isos, len(non_isos))
    unique_maxmisos = [maxmisos[i] for i in non_isos]
    duplicate_maxmisos = [maxmisos[i] for i in covered]
    logger.debug("unique_maxmisos", list(map(str, unique_maxmisos)))
    factors = [isos_size.get(i, 1) for i in non_isos]
    return unique_maxmisos, factors, duplicate_maxmisos


def get_func_query(session, stage, func_name):
    return Query(
        f"""MATCH p0=(n00:INSTR)-[r01:DFG]->(n01:INSTR)
        WHERE n00.func_name = '{func_name}'
        AND n00.session = "{session}"
        AND n00.stage = {stage}
        AND n01.name != "PHI" AND n01.name != "G_PHI"
        RETURN p0
        """
    )


def get_bbs_query(session, stage, func_name):
    return Query(
        f"""MATCH (n00:INSTR)
        WHERE n00.func_name = '{func_name}'
        AND n00.session = "{session}"
        AND n00.stage = {stage}
        RETURN n00.bb_id as bb_id, count(*) as num_instrs
        ORDER BY bb_id
        """
    )


def get_update_nodes_query(maxmiso_idx, maxmiso_nodes, factor=1):
    size = len(maxmiso_nodes)
    return Query(
        f"""
        MATCH (n:INSTR)
        WHERE id(n) IN {maxmiso_nodes}
        SET n.maxmiso_idx = {maxmiso_idx}, n.maxmiso_factor = {factor}, n.maxmiso_size = {size};
        """
    )


def query_candidates_from_db(
    sess: Session,
    workdir: Optional[Union[str, Path]] = None,
    label: Optional[str] = None,
    stage: int = 8,
    force: bool = False,
    progress: bool = False,
    query_config_yaml: Optional[Union[str, Path]] = None,
    limit_results: Optional[int] = None,
    # LIMIT_RESULTS: Optional[int] = 5000,
    # LIMIT_RESULTS: Optional[int] = 500,
    MIN_INPUTS: Optional[int] = 1,
    MAX_INPUTS: Optional[int] = 4,
    # MAX_INPUTS: Optional[int] = 3,
    MIN_OUTPUTS: Optional[int] = 1,
    MAX_OUTPUTS: Optional[int] = 2,
    # MAX_OUTPUTS: Optional[int] = 1,
    # TODO: implement topk!
    MAX_NODES: Optional[int] = int(1e3),
    # MAX_NODES: Optional[int] = int(5),
    MAX_ENC_FOOTPRINT: Optional[float] = 1.0,
    MAX_ENC_WEIGHT: Optional[float] = 1.0,
    MIN_ENC_BITS_LEFT: Optional[int] = 5,
    # MIN_NODES: Optional[Union[int, str]] = "auto",
    MIN_NODES: Optional[Union[int, str]] = None,
    MIN_PATH_LENGTH: Optional[int] = 1,
    # MAX_PATH_LENGTH = 3,
    MAX_PATH_LENGTH=9,
    MAX_PATH_WIDTH=2,
    # MAX_PATH_WIDTH = 4,
    INSTR_PREDICATES=511,  # ALL?
    IGNORE_NAMES=None,
    IGNORE_OP_TYPES=None,
    ALLOWED_ENC_SIZES=[32],
    min_iso_weight=0.05,
    scale_iso_weight: Union[bool, str] = "auto",
    MAX_LOADS=1,
    MAX_STORES=1,
    MAX_MEMS: Optional[int] = 0,  # TODO
    MAX_BRANCHES: Optional[int] = 1,
    xlen: Optional[int] = 32,  # TODO: do not hardcode
    ENABLE_VARIATION_REUSE_IO=False,
    # ENABLE_VARIATION_REUSE_IO=True,
    HALT_ON_ERROR: bool = True,
    sort_by: Optional[str] = "IsoWeight",
    topk: Optional[int] = None,
    partition_with_maxmiso: Union[str, bool] = "auto",
    enable_venn: bool = False,
):
    logger.info("Querying candidates from DB...")
    artifacts = sess.artifacts
    choices_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "choices")
    assert len(choices_artifacts) == 1
    choices_artifact = choices_artifacts[0]
    choices_df = choices_artifact.df
    logger.debug("choices_df", choices_df)
    if workdir is None:
        raise NotImplementedError("automatic workdir not supported yet!")
    if isinstance(workdir, str):
        workdir = Path(workdir)
    assert isinstance(workdir, Path)
    workdir = workdir.resolve()
    assert label is not None
    index_files = []
    # FUNC_ONLY = True
    FUNC_ONLY = False
    if FUNC_ONLY:
        # TODO: for real!
        funcs_df = choices_df.groupby("func_name", as_index=False)[["rel_weight", "num_instrs", "freq"]].sum()
        funcs_df["bb_name"] = None
        choices_df = funcs_df

    logger.debug("choices_df", choices_df)

    combined_query_metrics_df = pd.DataFrame()
    missing_bb_ids = set()
    for index, row in choices_df.iterrows():
        # logger.debug("index", index)
        # logger.debug("row", row)
        # if index != 1:
        #     continue
        # input(">")
        func_name = row["func_name"]
        bb_name = row["bb_name"]
        rel_weight = row["rel_weight"]
        num_instrs = row["num_instrs"]
        logger.debug("func_name", func_name)
        logger.debug("bb_name", bb_name)

        bbs_query = get_bbs_query(label, stage, func_name)

        memgraph_config = sess.config.memgraph
        memgraph_host = os.environ.get("MEMGRAPH_HOST")
        memgraph_port = os.environ.get("MEMGRAPH_PORT")
        user = ""
        password = ""
        if memgraph_config is not None:
            memgraph_host = memgraph_host or memgraph_config.hostname
            memgraph_port = memgraph_port or memgraph_config.port
            user = memgraph_config.user
            password = memgraph_config.password

        driver = GraphDatabase.driver(f"bolt://{memgraph_host}:{memgraph_port}", auth=(user, password))
        session = driver.session()
        try:
            func_query = get_func_query(label, stage, func_name)
            logger.debug("func_query", func_query)
            # print("func_query", func_query)
            logger.debug("bbs_query", bbs_query)
            func_results = session.run(func_query)
            func = memgraph_to_nx(func_results)
            logger.debug("func", func)
            # print("func", func)
            # input(">>")
            bbs_results = session.run(bbs_query)
            bbs_df = bbs_results.to_df()
        finally:
            session.close()
        # logger.debug("bbs_df", bbs_df)
        # input(">")
        bb_id = int(bb_name.split(".", 1)[1])
        assert len(bbs_df) > 0
        db_bb_ids = bbs_df["bb_id"].unique()
        if bb_id not in db_bb_ids:
            logger.warning(f"BB ID {bb_id} not found in DB!")
            missing_bb_ids.add((func_name, bb_id))
            continue
        maxmiso_idxs = []
        # TODO: skip maxmisos which are unsuitable?
        if isinstance(partition_with_maxmiso, str):
            # TODO: Handle 1/0/true/false/...
            # TODO: if numeric parse number!
            assert partition_with_maxmiso == "auto"
            num_instrs_threshold = 250
            partition_with_maxmiso = num_instrs > num_instrs_threshold
        if partition_with_maxmiso:
            bb_nodes = [node for node in func.nodes if func.nodes[node]["properties"]["bb_id"] == bb_id]
            # logger.debug("bb_nodes", bb_nodes)
            bb = func.subgraph(bb_nodes)
            # logger.debug("bb", bb)
            maxmisos = maxmiso_algo(bb)
            # logger.debug("maxmisos", maxmisos)
            # input(">")
            unique_maxmisos, factors, duplicate_maxmisos = get_unique_maxmisos(maxmisos)
            # TODO: write maxmisos to file?
            # maxmiso_node_ids = [[maxmiso.nodes[n]["key"] for n in maxmiso.nodes] for maxmiso in maxmisos]
            unique_maxmiso_node_ids = [[maxmiso.nodes[n]["key"] for n in maxmiso.nodes] for maxmiso in unique_maxmisos]
            duplicate_maxmiso_node_ids = [
                [maxmiso.nodes[n]["key"] for n in maxmiso.nodes] for maxmiso in duplicate_maxmisos
            ]
            remaining_nodes = set(bb_nodes)
            total_idx = 0
            session = driver.session()
            try:
                # TODO: make unique optional!
                # for i, node_ids in enumerate(maxmiso_node_ids):
                for i, node_ids in enumerate(duplicate_maxmiso_node_ids):
                    factor = 0
                    idx = -1
                    # Negative -> ignore
                    query = get_update_nodes_query(idx, node_ids, factor=0)
                    _ = session.run(query)
                    remaining_nodes -= set(node_ids)
                for i, node_ids in enumerate(unique_maxmiso_node_ids):
                    factor = factors[i]
                    query = get_update_nodes_query(total_idx, node_ids, factor=factor)
                    maxmiso_idxs.append(total_idx)
                    # logger.debug("query", query)
                    _ = session.run(query)
                    # logger.debug("results3", results3)
                    remaining_nodes -= set(node_ids)
                    total_idx += 1
                logger.debug("remaining_nodes", remaining_nodes)
                if len(remaining_nodes) > 0:
                    query = get_update_nodes_query(total_idx, list(remaining_nodes), factor=1)
                    _ = session.run(query)
                    maxmiso_idxs.append(total_idx)
                total_idx += 1
            finally:
                session.close()

        out_name = f"{func_name}_{bb_name}_0"
        out_dir = workdir / out_name
        logger.debug("out_dir", out_dir)
        out_dir.mkdir(exist_ok=True)
        index_file = out_dir / "index.yml"
        index_files.append(index_file)
        # TODO: to not hardcode this, call directly via python

        if scale_iso_weight is not None:
            if isinstance(scale_iso_weight, str):
                assert scale_iso_weight == "auto"
                scale_iso_weight = rel_weight
            assert isinstance(scale_iso_weight, float)
            assert scale_iso_weight > 0

            if min_iso_weight is not None:
                min_iso_weight *= scale_iso_weight
                # TODO: div by ?
                min_iso_weight = min(max(0.0, min_iso_weight), 1.0)

        # dynamically determine min_nodes
        min_nodes = MIN_NODES
        if isinstance(min_nodes, str):
            assert min_nodes == "auto"
            assert num_instrs != 0
            assert rel_weight > 0.0
            weight_per_instr = rel_weight / num_instrs
            assert weight_per_instr > 0.0
            min_nodes = None
            if min_iso_weight is not None:
                required_instrs = min_iso_weight / weight_per_instr
                # TODO: div by 2 to consider ISOs
                required_instrs = int(required_instrs + 0.5)  # TODO: round?
                required_instrs = max(1, required_instrs)
                if MAX_NODES is not None:
                    required_instrs = min(required_instrs, MAX_NODES)
                min_nodes = required_instrs

        args = [
            "python3",
            "-m",
            "tool.main",
        ]
        args += [
            *(["--progress"] if progress else []),
            *["--times"],
            *["--log", "info"],
            *(["--yaml", query_config_yaml] if query_config_yaml is not None else []),
            *["--session", label],
            *["--function", func_name],
            *(["--basic-block", bb_name] if bb_name is not None else []),
            *["--stage", str(stage)],
            *["--output-dir", out_dir],
            # *["--ignore-const-inputs"],
            *(["--limit-results", str(limit_results)] if limit_results is not None else []),
            *(["--min-inputs", str(MIN_INPUTS)] if MIN_INPUTS is not None else []),
            *(["--max-inputs", str(MAX_INPUTS)] if MAX_INPUTS is not None else []),
            *(["--min-outputs", str(MIN_OUTPUTS)] if MIN_OUTPUTS is not None else []),
            *(["--max-outputs", str(MAX_OUTPUTS)] if MAX_OUTPUTS is not None else []),
            *(["--max-nodes", str(MAX_NODES)] if MAX_NODES is not None else []),
            *(["--min-nodes", str(min_nodes)] if min_nodes is not None else []),
            *(["--max-loads", str(MAX_LOADS)] if MAX_LOADS is not None else []),
            *(["--max-loads", str(MAX_STORES)] if MAX_STORES is not None else []),
            *(["--max-mems", str(MAX_MEMS)] if MAX_MEMS is not None else []),
            *(["--max-branches", str(MAX_BRANCHES)] if MAX_BRANCHES is not None else []),
            *(["--max-enc-footprint", str(MAX_ENC_FOOTPRINT)] if MAX_ENC_FOOTPRINT is not None else []),
            *(["--max-enc-weight", str(MAX_ENC_WEIGHT)] if MAX_ENC_WEIGHT is not None else []),
            *(["--min-enc-bits-left", str(MIN_ENC_BITS_LEFT)] if MIN_ENC_BITS_LEFT is not None else []),
            *(["--min-path-length", str(MIN_PATH_LENGTH)] if MIN_PATH_LENGTH is not None else []),
            *(["--max-path-length", str(MAX_PATH_LENGTH)] if MAX_PATH_LENGTH is not None else []),
            *(["--max-path-width", str(MAX_PATH_WIDTH)] if MAX_PATH_WIDTH is not None else []),
            *(["--min-iso-weight", str(min_iso_weight)] if min_iso_weight is not None else []),
            *(["--instr-predicates", str(INSTR_PREDICATES)] if INSTR_PREDICATES is not None else []),
            *(["--ignore-names", IGNORE_NAMES] if IGNORE_NAMES is not None else []),
            *(["--ignore-op-types", str(IGNORE_OP_TYPES)] if IGNORE_OP_TYPES is not None else []),
            *(["--allowed-enc-sizes", " ".join(map(str, ALLOWED_ENC_SIZES))] if ALLOWED_ENC_SIZES is not None else []),
            *(["--xlen", str(xlen)] if xlen is not None else []),
            *(["--enable-variation-reuse-io"] if ENABLE_VARIATION_REUSE_IO else []),  # TODO: use FLT instead?
            *(["--halt-on-error"] if HALT_ON_ERROR else []),  # TODO: use FLT instead?
            *["--write-func"],
            # *["--write-func-fmt", WRITE_FUNC_FMT],
            # *["--write-func-flt", WRITE_FUNC_FLT],
            *["--write-sub"],
            # *["--write-sub-fmt", WRITE_SUB_FMT],
            # *["--write-sub-flt", WRITE_SUB_FLT],
            *["--write-io-sub"],
            # *["--write-io-sub-fmt", WRITE_IO_SUB_FMT],
            # *["--write-io-sub-flt", WRITE_IO_SUB_FLT],
            *["--write-tree"],
            # *["--write-tree-fmt", WRITE_TREE_FMT],
            # *["--write-treee-flt", WRITE_TREE_FLT],
            # *["--write-gen"],
            # *["--write-gen-fmt", WRITE_GEN_FMT],
            # *["--write-gen-flt", WRITE_GEN_FLT],
            *["--write-pie"],
            # *["--write-pie-fmt", WRITE_PIE_FMT],
            # *["--write-pie-flt", WRITE_PIE_FLT],
            *["--write-df"],
            # *["--write-df-fmt", WRITE_DF_FMT],
            # *["--write-df-flt", WRITE_DF_FLT],
            *["--write-index"],
            # *["--write-index-fmt", WRITE_INDEX_FMT],
            # *["--write-index-flt", WRITE_INDEX_FLT],
            *["--write-queries"],
            *["--write-query-metrics"],
        ]
        if partition_with_maxmiso and len(maxmiso_idxs) > 0:
            maxmiso_idxs_str = ",".join(map(str, maxmiso_idxs))
            args += ["--maxmisos", maxmiso_idxs_str]
        # args += ["--help"]
        logger.debug(">", " ".join(map(str, args)))
        subprocess.run(args, check=True)
        query_metrics_file = out_dir / "query_metrics.csv"
        query_metrics_df = pd.read_csv(query_metrics_file)
        query_metrics_df["func"] = func_name
        query_metrics_df["basic_block"] = bb_name
        combined_query_metrics_df = pd.concat([combined_query_metrics_df, query_metrics_df])

    # allow_missing_bb_id = False
    allow_missing_bb_id = True
    if len(missing_bb_ids) > 0:
        assert len(missing_bb_ids) < len(choices_df), "No matching BB IDs found in DB!"
        assert allow_missing_bb_id, f"Missing BB ID DB entries: {missing_bb_ids}"

    combined_query_metrics_file = workdir / "combined_query_metrics.csv"
    combined_query_metrics_df.to_csv(combined_query_metrics_file, index=False)
    combined_index_file = workdir / "combined_index.yml"
    combine_args = [
        "python3",
        "-m",
        "tool.combine_index",
        *index_files,
        "--drop-duplicates",
        "--drop-name-isos",  # NEW
        *(["--progress"] if progress else []),
        *(["--sort-by", sort_by] if sort_by is not None else []),
        *(["--topk", str(topk)] if topk is not None else []),
        "--out",
        combined_index_file,
    ]
    if enable_venn:
        if len(index_files) in [2, 3]:
            venn_diagram_file = workdir / "venn.jpg"
            combine_args += ["--venn", venn_diagram_file]
    sankey_diagram_file = workdir / "sankey.md"
    combine_args += ["--sankey", sankey_diagram_file]
    overlaps_file = workdir / "overlaps.csv"
    combine_args += ["--overlaps", overlaps_file]
    # logger.debug("combine_args", combine_args)
    subprocess.run(combine_args, check=True)

    # TODO: index and cdsl should use the same instruction names?
    # names = [f"CUSTOM{i}" for i, candidate in enumerate(combined_index_data["candidates"])]
    # num_fused_instrs = [
    #     candidate["properties"]["#Instrs"] for i, candidate in enumerate(combined_index_data["candidates"])
    # ]
    # names_df = pd.DataFrame({"instr": names, "num_fused_instrs": num_fused_instrs})
    # names_df["instr_lower"] = names_df["instr"].apply(lambda x: x.lower())
    names_csv = workdir / "names.csv"
    assign_names(combined_index_file, inplace=True, csv=names_csv)

    names_df = pd.read_csv(names_csv)
    if len(names_df) > 0:
        pdf_file = workdir / "all_io_subs.pdf"
        combine_pdfs(combined_index_file, output=pdf_file)

    # Extract names
    with open(combined_index_file, "r") as f:
        combined_index_data = yaml.safe_load(f)

    name2num_fused_instrs = {
        candidate["properties"]["InstrName"]: candidate["properties"]["#Instrs"]
        for i, candidate in enumerate(combined_index_data["candidates"])
    }
    names_df["num_fused_instrs"] = names_df["instr"].apply(lambda x: name2num_fused_instrs[x])
    # logger.debug("names_df", names_df)
    # input(">>>")
    attrs = {}
    ise_instrs_artifact = TableArtifact("ise_instrs", names_df, attrs=attrs)
    sess.add_artifact(ise_instrs_artifact, override=force)

    # gen_dir = Path(gen_dir) if gen_dir is not None else workdir / "gen"
    # gen_dir.mkdir(exist_ok=True)
    # generate_args = [
    #     combined_index_file,
    #     "--output",
    #     gen_dir,
    #     "--split",
    #     "--split-files",
    #     "--progress",
    #     "--inplace",  # TODO use gen/index.yml instead!
    # ]
    # generate_cdsl_args = [
    #     "python3",
    #     "-m",
    #     "tool.gen.cdsl",
    #     *generate_args,
    # ]
    # generate_flat_args = [
    #     "python3",
    #     "-m",
    #     "tool.gen.flat",
    #     *generate_args,
    # ]
    # generate_fuse_cdsl_args = [
    #     "python3",
    #     "-m",
    #     "tool.gen.fuse_cdsl",
    #     *generate_args,
    # ]
    # logger.debug("generate_cdsl_args", generate_cdsl_args)
    # logger.debug("generate_flat_args", generate_flat_args)
    # logger.debug("generate_fuse_cdsl_args", generate_fuse_cdsl_args)
    # subprocess.run(generate_cdsl_args, check=True)
    # subprocess.run(generate_flat_args, check=True)
    # subprocess.run(generate_fuse_cdsl_args, check=True)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    query_candidates_from_db(
        sess,
        workdir=args.workdir,
        label=args.label,
        stage=args.stage,
        force=args.force,
        progress=args.progress,
        query_config_yaml=args.query_config_yaml,
        limit_results=args.limit_results,
        xlen=args.xlen,
        min_iso_weight=args.min_iso_weight,
        scale_iso_weight=args.scale_iso_weight,
        sort_by=args.sort_by,
        topk=args.topk,
        partition_with_maxmiso=args.partition_with_maxmiso,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--workdir", type=str, default=None)
    parser.add_argument("--label", type=str, default=None)
    parser.add_argument("--progress", action="store_true", default=None)
    parser.add_argument("--stage", type=int, default=8)
    parser.add_argument("--limit-results", type=int, default=None)
    parser.add_argument("--query-config-yaml", type=str, default=None)
    parser.add_argument("--xlen", type=int, default=32)
    parser.add_argument("--min-iso-weight", type=float, default=None)
    parser.add_argument("--scale-iso-weight", type=str, default="auto")
    parser.add_argument("--topk", type=int, default=None)
    parser.add_argument("--sort-by", type=str, default=None)
    parser.add_argument("--partition-with-maxmiso", type=str, default="auto")
    # TODO: !
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import logging
import os.path as op
import subprocess
import sys
from pathlib import Path
from typing import Optional, Union
from collections import defaultdict

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import filter_artifacts

# Parts of the following code have been provided by MINRES Technologies.


def create_output(histogram, outputfile):
    """
    Write the histogram to the specified outputfile in lcov format
    """
    with open(outputfile, "w") as f:
        for file, lst in histogram.items():
            f.write(f"SF:{file}\n")
            executed = lst[0]
            functions = lst[1]
            # print("file", file)
            # print("executed", executed)
            # print("functions", functions)
            # input("*")
            for line, count in executed.items():
                if line in functions:
                    fcount = functions[line][0]
                    name = functions[line][1]
                    f.write(f"FN:{line},{name}\n")
                    f.write(f"FNDA:{fcount},{name}\n")
                f.write(f"DA:{line},{count}\n")
            f.write("end_of_record\n")


def generate_lcov_output(
    sess: Session,
    output: Optional[Union[str, Path]] = None,
    binary: bool = False,
    genhtml: Optional[Union[str, Path]] = None,
    force: bool = False,
):
    del force  # TODO: use
    artifacts = sess.artifacts

    func2pc_artifacts = filter_artifacts(artifacts, lambda x: x.name == "func2pc")
    assert len(func2pc_artifacts) == 1
    func2pc_artifact = func2pc_artifacts[0]
    func2pc_df = func2pc_artifact.df

    pcs_hist_artifacts = filter_artifacts(artifacts, lambda x: x.name == "pcs_hist")
    assert len(pcs_hist_artifacts) == 1
    pcs_hist_artifact = pcs_hist_artifacts[0]
    pcs_hist_df = pcs_hist_artifact.df

    pc2locs_artifacts = filter_artifacts(artifacts, lambda x: x.name == "pc2locs")
    assert len(pc2locs_artifacts) == 1
    pc2locs_artifact = pc2locs_artifacts[0]
    pc2locs_df = pc2locs_artifact.df

    # Build IntervalIndex
    if "start" not in func2pc_df.columns:
        func2pc_df[["start", "end"]] = func2pc_df["pc_range"].apply(pd.Series)
    intervals = pd.IntervalIndex.from_arrays(func2pc_df["start"], func2pc_df["end"], closed="both")
    func_names = func2pc_df["func"].values

    # Map function to a PC
    def get_func_name(pc):
        matches = intervals.contains(pc)
        if matches.any():
            return func_names[matches.argmax()]  # first matching interval
        return None

    # Annotate pcs_hist_df
    func_agg = "all"
    if func_agg == "all":
        pcs_hist_df["func_name"] = pcs_hist_df["pc"].apply(get_func_name)
    elif func_agg == "first":
        raise NotImplementedError(f"func_agg: {func_agg}")
    else:
        raise ValueError(f"Invalid func_agg: {func_agg}")

    pc2locs_df["locs"] = pc2locs_df["locs"].apply(list)

    # Extract the loc with the smallest line number
    def pick_min_loc(locs):
        # locs is a list like ["src/foo.c:260", "src/foo.c:389"]
        return min(locs, key=lambda x: int(x.split(":")[1]))

    pc2locs_df["locs"] = pc2locs_df["locs"].apply(pick_min_loc)

    merged_df = pcs_hist_df.merge(pc2locs_df, on="pc", how="left")
    # Sort by function and pc
    merged_df = merged_df.sort_values(["func_name", "pc"]).reset_index(drop=True)
    # Forward-fill locs *within each function*
    merged_df["locs"] = merged_df.groupby("func_name")["locs"].ffill()

    # Now group by locs and sum counts
    locs_hist_df = (
        merged_df.groupby(["locs", "func_name"], dropna=False)
        .agg(count=("count", "sum"), rel_count=("rel_count", "sum"))
        .reset_index()
    )
    # print("locs_hist_df", locs_hist_df)

    # Sort by count if desired
    locs_hist_df = locs_hist_df.sort_values("count", ascending=False)
    # If locs are strings like "src/core_list_join.c:260"
    # Split into file and line
    locs_hist_df[["file", "line"]] = locs_hist_df["locs"].str.split(":", n=1, expand=True)
    locs_hist_df = locs_hist_df.drop(columns=["locs", "rel_count"])
    locs_hist_df["line"] = locs_hist_df["line"].fillna(0)
    locs_hist_df["line"] = locs_hist_df["line"].astype(int)

    # Structure: {filename: [executed_dict, functions_dict]}
    # executed_dict: line → count
    # functions_dict: line → [count, name]
    histogram = defaultdict(lambda: [defaultdict(int), defaultdict(lambda: [0, ""])])
    for _, row in locs_hist_df.iterrows():
        file = row["file"]
        line = row["line"]
        count = row["count"]
        func_name = row["func_name"]

        # Increment executed count
        if binary:
            histogram[file][0][line] = 1
        else:
            histogram[file][0][line] += count

        # If function info is available
        if func_name is not None:
            if binary:
                histogram[file][1][line][0] = 1
            else:
                histogram[file][1][line][0] += count
            histogram[file][1][line][1] = func_name  # store function name

    if output is None:
        profile_dir = sess.directory / "profile"
        profile_dir.mkdir(exist_ok=True)
        out_name = "coverage.info"
        output = profile_dir / out_name
        print("genhtml", genhtml)
        if genhtml is None:  # TODO: make optional?
            genhtml = profile_dir / "html"

    create_output(
        histogram,
        output,
    )

    if binary:
        logging.info("Displaying binary coverage")

    if op.getsize("coverage.info") == 0:
        logging.critical("Something went wrong, couldn't generate output")
        sys.exit(1)

    if genhtml:
        logging.info(f"Output directory set to {genhtml}")
        try:
            subprocess.run(
                [
                    "genhtml",
                    "-o",
                    genhtml,
                    "-q",
                    output,
                    "--ignore-errors",
                    "source",
                ],
                check=True,
            )
        except FileNotFoundError:
            logging.error('Could not find "genhtml", is the lcov package installed?')
            sys.exit(1)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    generate_lcov_output(
        sess,
        output=args.output,
        genhtml=args.genhtml,
        binary=args.binary,
        force=args.force,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument(
        "--genhtml",
        metavar="<directory>",
        nargs="?",
        default=None,  # TODO: ?
        help=(
            "calls the genhtml command from lcov, will generate a directory for the "
            "generated output if not specified otherwise"
        ),
    )
    parser.add_argument(
        "--binary",
        "-b",
        action="store_true",
        default=False,
        help="toggle boolean coverage",
    )
    parser.add_argument(
        "--output",
        "-o",
        type=str,
        metavar="<outputfile>",
        nargs="?",
        default=None,
        help="specify the putput file",
    )
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

from isaac_toolkit.session import Session

# from isaac_toolkit.logging import get_logger, set_log_level
from isaac_toolkit.analysis.static.dwarf import analyze_dwarf

# from isaac_toolkit.analysis.static.llvm_bbs import analyze_llvm_bbs
from isaac_toolkit.analysis.static.mem_footprint import analyze_mem_footprint
from isaac_toolkit.analysis.static.linker_map import analyze_linker_map
from isaac_toolkit.analysis.static.histogram.disass_instr import (
    create_disass_instr_hist,
)
from isaac_toolkit.analysis.static.histogram.disass_opcode import (
    create_disass_opcode_hist,
)

# from isaac_toolkit.analysis.dynamic.trace.trunc_trace import trunc_trace
# from isaac_toolkit.analysis.dynamic.trace.instr_operands import analyze_instr_operands
from isaac_toolkit.analysis.dynamic.histogram.opcode import create_opcode_hist
from isaac_toolkit.analysis.dynamic.histogram.pc import create_pc_hist

# from isaac_toolkit.analysis.dynamic.histogram.opcode_per_llvm_bb import (
#     create_opcode_per_llvm_bb_hist,
# )
from isaac_toolkit.analysis.dynamic.histogram.instr import create_instr_hist

# from isaac_toolkit.analysis.dynamic.trace.basic_blocks import analyze_basic_blocks
from isaac_toolkit.analysis.dynamic.trace.trace_bbs import collect_trace_bbs

# from isaac_toolkit.analysis.dynamic.trace.map_llvm_bbs_new import map_llvm_bbs_new
from isaac_toolkit.analysis.dynamic.trace.track_used_functions import (
    track_unused_functions,
)

# logger = get_logger()
import logging

logger = logging.getLogger()


def analyze_artifacts(sess: Session, force: bool = False, progress: bool = False):
    logger.info("Analyzing RVF Demo artifacts...")
    analyze_dwarf(sess, force=force)
    # analyze_llvm_bbs(sess, force=force)
    analyze_mem_footprint(sess, force=force)
    analyze_linker_map(sess, force=force)
    # trunc_trace(sess, start_func="mlonmcu_run", force=force)
    # trunc_trace(sess, end_func="stop_bench", force=force)
    # analyze_instr_operands(sess, force=force)  # TODO: separate df and detect if available
    create_opcode_hist(sess, force=force)
    # create_opcode_per_llvm_bb_hist(sess, force=force)
    create_instr_hist(sess, force=force)
    create_pc_hist(sess, force=force)
    create_disass_instr_hist(sess, force=force)
    create_disass_opcode_hist(sess, force=force)
    # analyze_basic_blocks(sess, force=force)
    collect_trace_bbs(sess, force=force)
    # map_llvm_bbs_new(sess, force=force)
    track_unused_functions(sess, force=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    # set_log_level(console_level=args.log, file_level=args.log)
    analyze_artifacts(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def map_llvm_bbs(sess: Session, force: bool = False):
    logger.info("Mapping LLVM BBs (old)...")
    artifacts = sess.artifacts
    # print("artifacts", artifacts)
    elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    # print("elf_artifacts", elf_artifacts)
    assert len(elf_artifacts) == 1
    elf_artifact = elf_artifacts[0]
    trace_pc2bb_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "pc2bb"
    )  # TODO: optional or different pass
    assert len(trace_pc2bb_artifacts) == 1
    trace_pc2bb_artifact = trace_pc2bb_artifacts[0]
    trace_pc2bb_df = trace_pc2bb_artifact.df
    # print("trace_pc2bb_df", trace_pc2bb_df)
    llvm_bbs_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "llvm_bbs"
    )  # TODO: optional or different pass
    assert len(llvm_bbs_artifacts) == 1
    llvm_bbs_artifact = llvm_bbs_artifacts[0]
    llvm_bbs_df = llvm_bbs_artifact.df.copy()
    llvm_bbs_df[["start", "end"]] = llvm_bbs_df["pcs"].apply(pd.Series)
    # print("llvm_bbs_df", llvm_bbs_df)
    for index, row in llvm_bbs_df.sort_values("start").iterrows():
        start = row["start"]
        end = row["end"]

        def find_matching_bb(pc):
            # print("find_matching_bb", pc)
            matches = trace_pc2bb_df.where(lambda x: x["start"] < pc).dropna()
            # print("m1", matches)
            matches = matches.where(lambda x: pc < x["end"]).dropna()
            # print("m2", matches)
            # print("matches", matches)
            if len(matches) == 0:
                return None
            assert len(matches) == 1
            return matches.iloc[0]

        matching_row_start = [find_matching_bb(start)]

        def split_trace_bb(idx, row, start=None, end=None):
            # print("split_trace_bb", idx, row, start, end)
            # idx = row.index[0]
            if start is not None:
                # print(f"SPLIT {row.index} @ {start} (start)")
                pass
                # orig_start = row["start"]
                # input(">")
            if end is not None:
                pass
                # print(f"SPLIT {row.index} @ {end} (end)")
                # input(">")

        # print("matching_row_start", matching_row_start)
        for row in matching_row_start:
            if row is None:
                continue
            # if row["start"] == start:
            #     continue
            split_trace_bb(index, row, start=start)
        matching_row_end = [find_matching_bb(end)]
        # print("matching_row_end", matching_row_end)
        for row in matching_row_end:
            if row is None:
                continue
            # if row["end"] == end:
            #     continue
            split_trace_bb(index, row, end=end)

    # input("!")
    def helper(x):
        # print("x", x)
        ret = set()
        func_names = x["func_name"]
        start = x["start"]
        end = x["end"]
        for func_name in func_names:
            func_matches = llvm_bbs_df[llvm_bbs_df["func_name"] == func_name]
            # print("func_matches", func_matches)
            start_matches = func_matches.where(lambda x: x["start"] <= start).dropna()
            # print("start_matches", start_matches)
            end_matches = start_matches.where(lambda x: x["end"] >= end).dropna()
            # print("end_matches", end_matches)
            for bb_name in end_matches["bb_name"]:
                ret.add(bb_name)
        # input("b")
        # x["test"] = 42
        return ret

    trace_pc2bb_df["llvm_bbs"] = trace_pc2bb_df[["func_name", "start", "end"]].apply(lambda x: helper(x), axis=1)
    # print("trace_pc2bb_df new", trace_pc2bb_df)
    remain = trace_pc2bb_df[trace_pc2bb_df["llvm_bbs"].map(len) == 0]

    # print("remain", remain)
    def helper2(x):
        # print("x", x)
        ret = set()
        func_names = x["func_name"]
        start = x["start"]
        end = x["end"]
        for func_name in func_names:
            func_matches = llvm_bbs_df[llvm_bbs_df["func_name"] == func_name]
            # print("func_matches", func_matches)
            start_matches = func_matches.where(lambda x: x["start"] <= start).dropna()
            # print("start_matches", start_matches)
            end_matches = start_matches.where(lambda x: x["end"] >= end).dropna()
            # print("end_matches", end_matches)
            for bb_name in end_matches["bb_name"]:
                ret.add(bb_name)
        # input("b")
        # x["test"] = 42
        return ret

    remain["llvm_bbs"] = remain[["func_name", "start", "end"]].apply(lambda x: helper2(x), axis=1)
    # print("remain new", remain)
    # input("a1")

    attrs = {
        "elf_file": elf_artifact.name,
        "kind": "mapping",
        "by": __name__,
    }

    artifact = TableArtifact("pc2bb_llvm", trace_pc2bb_df, attrs=attrs)
    # print("artifact", artifact)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    map_llvm_bbs(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import argparse
from pathlib import Path
from collections import defaultdict

import yaml


def filter_index(
    index,
    output=None,
    inplace=None,
    sankey=None,
    min_seal5_score=None,
    min_util_score=None,
    min_runtime_reduction_rel=None,
    min_code_size_reduction_rel=None,
    min_estimated_reduction=None,
):
    with open(index, "r") as f:
        combined_index_data = yaml.safe_load(f)
    candidates = combined_index_data["candidates"]
    num_candidates = len(candidates)

    to_drop = set()
    reasons_dropped = defaultdict(set)
    for i, candidate in enumerate(candidates):
        metrics = candidate.get("metrics", {})
        if min_seal5_score is not None:
            seal5_score = metrics.get("seal5_score")
            if seal5_score is not None:
                if seal5_score < min_seal5_score:
                    to_drop.add(i)
                    reasons_dropped["min_seal5_score"].add(i)
                    continue
        if min_util_score is not None:
            util_score = metrics.get("util_score")
            if util_score is not None:
                if util_score < min_util_score:
                    to_drop.add(i)
                    reasons_dropped["min_util_score"].add(i)
                    continue
        if min_runtime_reduction_rel is not None:
            runtime_reduction_rel = metrics.get("runtime_reduction_rel")
            if runtime_reduction_rel is not None:
                if runtime_reduction_rel < min_runtime_reduction_rel:
                    to_drop.add(i)
                    reasons_dropped["min_runtime_reduction_rel"].add(i)
                    continue
        if min_code_size_reduction_rel is not None:
            code_size_reduction_rel = metrics.get("code_size_reduction_rel")
            if code_size_reduction_rel is not None:
                if code_size_reduction_rel < min_code_size_reduction_rel:
                    to_drop.add(i)
                    reasons_dropped["min_code_size_reduction_rel"].add(i)
                    continue
        if min_estimated_reduction is not None:
            estimated_reduction_rel = metrics.get("estimated_reduction_rel")
            if estimated_reduction_rel is not None:
                if estimated_reduction_rel < min_estimated_reduction:
                    to_drop.add(i)
                    reasons_dropped["min_estimated_reduction"].add(i)
                    continue
    num_drop = len(to_drop)
    num_keep = num_candidates - num_drop
    # TODO: logging
    # TODO: assign new names?
    candidates = [x for i, x in enumerate(candidates) if i not in to_drop]
    combined_index_data["candidates"] = candidates

    if inplace:
        assert output is None
        out_file = index
    else:
        assert output is not None
        out_file = output

    with open(out_file, "w") as f:
        yaml.dump(combined_index_data, f)

    if sankey is not None:
        # logger.info("Exporting sankey diagram...")
        fmt = Path(sankey).suffix
        assert fmt in [".md"]
        content = """
```mermaid
---
config:
  sankey:
    showValues: true
---
sankey-beta

%% source,target,value
"""
        content += f"Candidates,Kept,{num_keep}\n"
        for reason, reason_dropped in reasons_dropped.items():
            reason_num_dropped = len(reason_dropped)
            content += f"Candidates,Filtered({reason}),{reason_num_dropped}\n"
        content += """
```
"""
        with open(sankey, "w") as f:
            f.write(content)


def handle(args):
    filter_index(
        args.index,
        output=args.output,
        inplace=args.inplace,
        sankey=args.sankey,
        min_seal5_score=args.min_seal5_score,
        min_util_score=args.min_util_score,
        min_runtime_reduction_rel=args.min_runtime_reduction_rel,
        min_code_size_reduction_rel=args.min_code_size_reduction_rel,
        min_estimated_reduction=args.min_estimated_reduction,
    )


def get_parser():
    parser = argparse.ArgumentParser(description="TODO")
    parser.add_argument("index", help="Index yaml file")
    parser.add_argument("-o", "--output", default=None, help="Output yaml file")
    parser.add_argument("--inplace", action="store_true", help="TODO")
    parser.add_argument("--sankey", default=None, help="TODO")
    parser.add_argument("--min-seal5-score", type=float, default=None, help="TODO")
    parser.add_argument("--min-util-score", type=float, default=None, help="TODO")
    parser.add_argument("--min-runtime-reduction-rel", type=float, default=None, help="TODO")
    parser.add_argument("--min-code-size-reduction-rel", type=float, default=None, help="TODO")
    parser.add_argument("--min-estimated-reduction", type=float, default=None, help="TODO")  # TODO: rename
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import logging
import argparse
import itertools
from pathlib import Path
from typing import Union, Optional
from concurrent.futures import ProcessPoolExecutor

# import multiprocessing


# from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed

import pandas as pd
import numpy as np

# from tqdm import tqdm

# from isaac_toolkit.arch.riscv import riscv_branch_instrs, riscv_return_instrs

from capstone import Cs, CS_ARCH_RISCV, CS_MODE_RISCV64, CS_MODE_RISCV32, CS_MODE_RISCVC
from elftools.elf.elffile import ELFFile
from elftools.elf.constants import SH_FLAGS

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts, InstrTraceArtifact


logging.basicConfig(level=logging.DEBUG)  # TODO
logger = logging.getLogger(__name__)


class ELFInstructionFetcher:
    def __init__(self, elf_path):
        self.sections = []
        with open(elf_path, "rb") as f:
            elf = ELFFile(f)
            for section in elf.iter_sections():
                flags = section["sh_flags"]
                if flags & SH_FLAGS.SHF_EXECINSTR:
                    base = section["sh_addr"]
                    size = section["sh_size"]
                    data = section.data()
                    self.sections.append((base, base + size, data))

    def read_word_at_pc(self, pc, size=4):
        """Fetch `size` bytes at `pc` from any executable section."""
        for start, end, data in self.sections:
            if start <= pc < end:
                offset = pc - start
                if offset + size > len(data):
                    return None  # avoid out-of-bounds read
                return int.from_bytes(data[offset : offset + size], "little")
        return None  # pc not found


def disassemble_word(md, pc, word, size=4, endian="little", operands: bool = False):
    # print("disassemble_word", pc, word, size)
    try:
        code = word.to_bytes(size, endian)
        # print("code", code)
        insn = next(md.disasm(code, pc))
        # print("insn", insn)
        # input("%1")
        if operands:
            return insn.mnemonic, insn.op_str
        return insn.mnemonic
    except Exception:
        if operands:
            return "unknown", ""
        return "unknown"


def get_disassembler(elf_path):
    with open(elf_path, "rb") as f:
        elf = ELFFile(f)
    xlen = elf.elfclass
    mode = CS_MODE_RISCV32 if xlen == 32 else CS_MODE_RISCV64

    md = Cs(CS_ARCH_RISCV, mode | CS_MODE_RISCVC)
    md.detail = False
    return md


def build_entries(rows, elf_path):
    fetcher = ELFInstructionFetcher(elf_path)
    md = get_disassembler(elf_path)

    out = []
    for pc, size in rows:
        bytecode = fetcher.read_word_at_pc(int(pc), size=int(size))
        instr = disassemble_word(md, int(pc), int(bytecode), size=int(size), operands=False)
        out.append(((pc, size), (bytecode, instr)))
    return out


def worker(row_chunk, elf_path):
    fetcher = ELFInstructionFetcher(elf_path)
    md = get_disassembler(elf_path)
    out = {}
    for pc, size in row_chunk:
        bytecode = fetcher.read_word_at_pc(pc, size=size)
        instr = disassemble_word(md, pc, bytecode, size=size, operands=False)
        out[(pc, size)] = (bytecode, instr)
    return out


def parallel_dict_builder(rows, elf_path, nprocs=16):
    chunks = np.array_split(rows, nprocs)
    with ProcessPoolExecutor(max_workers=nprocs) as pool:
        results = pool.map(worker, chunks, itertools.repeat(elf_path))
    merged = {}
    for r in results:
        merged.update(r)
    return merged


def normalize_trace(
    sess: Session,
    output: Optional[Union[str, Path]] = None,
    force: bool = False,
):
    artifacts = sess.artifacts
    elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    elf_artifact = None
    if len(elf_artifacts) > 0:
        assert len(elf_artifacts) == 1
        elf_artifact = elf_artifacts[0]

    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]
    trace_df = trace_artifact.df
    # print("trace_df", len(trace_df), trace_df.memory_usage(deep=True), trace_df.memory_usage(deep=True).sum())

    # pc2bytecode = {}

    # def lookup_bytecode(pc, size):
    #     pc = int(pc)
    #     bytecode = pc2bytecode.get(pc)
    #     if bytecode is None:  # not in cache
    #         bytecode = fetcher.read_word_at_pc(pc, size=size)
    #         assert bytecode is not None
    #         pc2bytecode[pc] = bytecode
    #     return bytecode
    #
    # bytecode2name = {}
    #
    # def lookup_name(bytecode, pc=None, size=None):
    #     bytecode = int(bytecode)
    #     name = bytecode2name.get(bytecode)
    #     if name is None:  # not in cache
    #         assert pc is not None
    #         assert size is not None
    #         name = disassemble_word(pc, bytecode, size=size)
    #         assert name is not None
    #         bytecode2name[bytecode] = name
    #     # print("name", name)
    #     # input("?")
    #     return name

    md = None

    def disassemble_row(row):
        return disassemble_word(md, int(row["pc"]), int(row["bytecode"]), size=int(row["size"]), operands=False)

    cols = trace_df.columns
    # print("cols", cols)
    has_pc = "pc" in cols
    has_bytecode = "bytecode" in cols
    has_size = "size" in cols
    has_instr = "instr" in cols
    # print("has", has_pc, has_bytecode, has_size, has_instr)
    assert has_pc, "Missing required column in trace_df: pc"

    inplace = True
    if not inplace:
        trace_df = trace_df.copy()

    if not has_size:
        logger.info("Filling missing column: size")
        # TODO: detect from major opcode or elf?
        raise NotImplementedError("size")
        has_size = True

    if not has_instr:
        assert elf_artifact is not None
        md = get_disassembler(elf_artifact.path)

    if not has_bytecode:
        logger.info("Filling missing column: bytecode")
        assert elf_artifact is not None
        fetcher = ELFInstructionFetcher(elf_artifact.path)
        assert has_size
        # unique_pcs = trace_df["pc"].unique()
        unique_pc_size = trace_df.drop_duplicates(subset=["pc"])[["pc", "size"]]
        unique_pc_size["bytecode"] = unique_pc_size.apply(
            lambda x: fetcher.read_word_at_pc(x["pc"], size=x["size"]), axis=1
        )
        # print("unique_pc_size", unique_pc_size)
        assert all(unique_pc_size["bytecode"].notna()), "Unable to get all bytecodes"
        unique_pc_size["bytecode"] = pd.to_numeric(unique_pc_size["bytecode"], downcast="unsigned")
        unique_pc_size["bytecode"] = unique_pc_size["bytecode"].astype("category")
        if not has_instr:
            logger.info("Filling missing column: instr")
            unique_pc_size["instr"] = unique_pc_size.apply(disassemble_row, axis=1)
            unique_pc_size["instr"] = unique_pc_size["instr"].astype("category")
            has_instr = True
        # pc_size_to_bytecode = dict(zip(zip(unique_pc_size["pc"], unique_pc_size["size"]), unique_pc_size["bytecode"]))
        # pc_size_to_instr = dict(zip(zip(unique_pc_size["pc"], unique_pc_size["size"]), unique_pc_size["instr"]))
        # trace_df["bytecode"] = list(map(lambda k: pc_size_to_bytecode[k], zip(trace_df["pc"], trace_df["size"])))
        # trace_df["instr"] = list(map(lambda k: pc_size_to_instr[k], zip(trace_df["pc"], trace_df["size"])))
        unique_pc_size.drop(columns=["size"], inplace=True)
        trace_df = trace_df.merge(unique_pc_size, on=["pc"], how="left")
        has_bytecode = True
        # rows = list(zip(unique_pc_size["pc"], unique_pc_size["size"]))
        # pc_size_dict = parallel_dict_builder(rows, elf_artifact.path, nprocs=16)
        # print("pc_size_dict", pc_size_dict)

        # trace_df["bytecode"] = [pc_size_dict[(pc, size)][0] for pc, size in zip(trace_df["pc"], trace_df["size"])]
        # trace_df["instr"] = [pc_size_dict[(pc, size)][1] for pc, size in zip(trace_df["pc"], trace_df["size"])]

    if not has_instr:

        logger.info("Filling missing column: instr")
        unique_pc_size = trace_df[["pc", "size"]].drop_duplicates()
        unique_pc_size["instr"] = unique_pc_size.apply(disassemble_row, axis=1)
        unique_pc_size["instr"] = unique_pc_size["instr"].astype("category")
        trace_df = trace_df.merge(unique_pc_size, on=["pc", "size"], how="left")
        has_instr = True

    # TODO: normalize dtypes! (check overhead)
    from pandas.api.types import is_unsigned_integer_dtype

    if not is_unsigned_integer_dtype(trace_df["pc"].dtype):
        logger.info("Fixing dtype of column: pc")
        trace_df["pc"] = pd.to_numeric(trace_df["pc"], downcast="unsigned")

    if not isinstance(trace_df["instr"].dtype, pd.CategoricalDtype):
        logger.info("Fixing dtype of column: instr")
        trace_df["instr"] = trace_df["instr"].astype("category")

    if not isinstance(trace_df["size"].dtype, pd.CategoricalDtype):
        logger.info("Fixing dtype of column: size")
        trace_df["size"] = trace_df["size"].astype("category")

    # TODO: handle is_branch/is_return

    if not isinstance(trace_df["bytecode"].dtype, pd.CategoricalDtype):
        logger.info("Fixing dtype of column: bytecode")
        if not is_unsigned_integer_dtype(trace_df["bytecode"].dtype):
            trace_df["bytecode"] = pd.to_numeric(trace_df["bytecode"], downcast="unsigned")
        trace_df["bytecode"] = trace_df["bytecode"].astype("category")

    # TODO: check if saved?
    # print(
    #     "trace_df",
    #     len(trace_df),
    #     trace_df.memory_usage(deep=True),
    #     trace_df.memory_usage(deep=True).sum(),
    #     trace_df.dtypes,
    # )

    if not inplace:
        artifact = InstrTraceArtifact(trace_artifact.name, trace_df, attrs=trace_artifact.attrs)
        sess.add_artifact(artifact, override=force)
    else:
        trace_artifact._data = trace_df
        trace_artifact.update()  # TODO: pass new data via arg


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    normalize_trace(
        sess,
        force=args.force,  # TODO: check if update requires force?
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# import time
import sys
import argparse
from typing import List
from pathlib import Path

import pandas as pd
import numpy as np

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import TraceArtifact


# TODO: logger


def load_perf_trace(sess: Session, input_files: List[Path], force: bool = False):
    assert len(input_files) > 0
    name = input_files[0].name
    # sort input files by name
    sorted_files = sorted(input_files, key=lambda x: x.name)
    dfs = list(
        map(
            lambda x: pd.read_csv(x, sep=r"\s*\|\s*", engine="python")
            .dropna(axis=1)
            .replace("------------------", np.nan)
            # .map(lambda y: int(y, 0) if not pd.isna(y) else y),  # Python 3.10
            .applymap(lambda y: int(y, 0) if not pd.isna(y) else y),  # Python 3.8
            sorted_files,
        )
        # TODO: rewrite to be less hacky and more efficient
    )
    df = pd.concat(dfs, axis=0)
    df.reset_index(inplace=True, drop=True)
    # df["instr"] = df["instr"].astype("category")
    # df["size"] = df["size"].astype("category")

    attrs = {
        "simulator": "etiss",
        "cpu_arch": "unknown",
        "kind": "perf_trace",
        "by": "isaac_toolkit.frontend.perf_trace.etiss_perf",
    }
    artifact = TraceArtifact(name, df, attrs=attrs)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    input_files = list(map(Path, args.files))
    load_perf_trace(sess, input_files, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("files", nargs="+")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
import logging
from pathlib import Path

import pandas as pd
import matplotlib.pyplot as plt

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()

logging.getLogger("matplotlib.font_manager").setLevel(logging.WARNING)


# TODO: share with other pie scripts
def plot_pie_data(series, y, threshold: float = 0.1, title: str = "Pie Chart", legend: bool = True):

    def make_autopct(values):
        def my_autopct(pct):
            total = sum(values)
            val = int(round(pct * total / 100.0))
            return "{p:.2f}%\n({v:d})".format(p=pct, v=val)

        return my_autopct

    plot = series.plot.pie(
        y=y,
        autopct="%1.1f%%",
        # autopct=make_autopct(series[y].values),
        # legend=legend,
        title=title,
        labeldistance=1 if not legend else None,
    )
    if legend:
        # Matplotlibs hides legend labels starting with an '_'...
        labels = [f" {x}" if str(x).startswith("_") else x for x in series.index]
        plt.legend(labels, loc="center left", bbox_to_anchor=(1.0, 0.5))
    return plot


# TODO: share with other pie scripts
def generate_pie_data(df, x: str, y: str, topk: int = 9):
    ret = df.copy()
    ret.set_index(x, inplace=True)
    ret.sort_values(y, inplace=True, ascending=False)
    a = ret.iloc[:topk]
    b = ret.iloc[topk:].agg(others=(y, "sum"))

    ret = pd.concat([a, b])
    ret = ret[y]

    return ret


# TODO: share with other pie scripts
def agg_library_runtime(runtime_df, symbol_map_df, by: str = "library", col: str = "rel_weight"):
    ret = runtime_df.set_index("func_name").join(symbol_map_df.set_index("symbol"), how="left")
    ret = ret[[by, col]]
    ret = ret.groupby(by, as_index=False, dropna=False).sum()
    return ret


def create_runtime_pie_plots(
    sess: Session,
    threshold: float = 0.05,
    topk: int = 9,
    fmt: str = "jpg",
    legend: bool = True,
    force: bool = False,
):
    logger.info("Visualizing runtime via pie chars...")
    artifacts = sess.artifacts
    # TODO: allow missing files!
    pc2bb_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "pc2bb")
    assert len(pc2bb_artifacts) == 1
    pc2bb_artifact = pc2bb_artifacts[0]
    pc2bb_df = pc2bb_artifact.df

    symbol_map_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "symbol_map"
    )
    symbol_map_df = None
    if len(symbol_map_artifacts) > 0:
        assert len(symbol_map_artifacts) == 1
        symbol_map_artifact = symbol_map_artifacts[0]
        symbol_map_df = symbol_map_artifact.df

    instrs_hist_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "instrs_hist"
    )
    instrs_hist_df = None
    if len(instrs_hist_artifacts) > 0:
        assert len(instrs_hist_artifacts) == 1
        instrs_hist_artifact = instrs_hist_artifacts[0]
        instrs_hist_df = instrs_hist_artifact.df

    opcodes_hist_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "opcodes_hist"
    )
    opcodes_hist_df = None
    if len(opcodes_hist_artifacts) > 0:
        assert len(opcodes_hist_artifacts) == 1
        opcodes_hist_artifact = opcodes_hist_artifacts[0]
        opcodes_hist_df = opcodes_hist_artifact.df

    llvm_bbs_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "llvm_bbs_new"
    )
    llvm_bbs_df = None
    if len(llvm_bbs_artifacts) > 0:
        assert len(llvm_bbs_artifacts) == 1
        llvm_bbs_artifact = llvm_bbs_artifacts[0]
        llvm_bbs_df = llvm_bbs_artifact.df.copy()

    plots_dir = sess.directory / "plots"
    plots_dir.mkdir(exist_ok=True)
    # TODO: use threshold

    if pc2bb_df is not None:
        runtime_df = pc2bb_df.copy()

        def helper(x):
            if x is None:
                return "?"
            if isinstance(x, set):
                # assert len(x) == 1
                assert len(x) > 0
                # Only pick first element if alias exists
                return list(x)[0]
            return x

        runtime_df["func_name"] = runtime_df["func_name"].apply(helper)
        runtime_df = runtime_df[["func_name", "rel_weight"]]
        runtime_df = runtime_df.groupby("func_name", as_index=False, dropna=False).sum()
        runtime_per_func_data = generate_pie_data(runtime_df, x="func_name", y="rel_weight", topk=topk)
        runtime_per_func_plot = plot_pie_data(
            runtime_per_func_data,
            "rel_weight",
            threshold=threshold,
            legend=legend,
            title="Runtime per Func",
        )
        runtime_per_func_plot_file = plots_dir / f"runtime_per_func.{fmt}"
        if runtime_per_func_plot_file.is_file():
            assert force, f"File already exists: {runtime_per_func_plot_file}"
        runtime_per_func_plot.get_figure().savefig(
            runtime_per_func_plot_file,
            bbox_inches="tight",
            dpi=300,
        )
        plt.close()
        if llvm_bbs_df is not None:
            llvm_bbs_df["func_bb"] = llvm_bbs_df["func_name"] + "-" + llvm_bbs_df["bb_name"]
            runtime_per_llvm_bb_data = generate_pie_data(llvm_bbs_df, x="func_bb", y="rel_weight", topk=topk)
            runtime_per_llvm_bb_plot = plot_pie_data(
                runtime_per_llvm_bb_data,
                "rel_weight",
                threshold=threshold,
                legend=legend,
                title="Runtime per LLVM BB",
            )
            runtime_per_llvm_bb_plot_file = plots_dir / f"runtime_per_llvm_bb.{fmt}"
            if runtime_per_llvm_bb_plot_file.is_file():
                assert force, f"File already exists: {runtime_per_llvm_bb_plot_file}"
            runtime_per_llvm_bb_plot.get_figure().savefig(
                runtime_per_llvm_bb_plot_file,
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()
        if symbol_map_df is not None:
            # library
            library_runtime_df = agg_library_runtime(runtime_df, symbol_map_df, by="library", col="rel_weight")
            runtime_per_library_data = generate_pie_data(library_runtime_df, x="library", y="rel_weight", topk=topk)
            runtime_per_library_plot = plot_pie_data(
                runtime_per_library_data,
                "rel_weight",
                threshold=threshold,
                legend=legend,
                title="Runtime per Library",
            )
            runtime_per_library_plot_file = plots_dir / f"runtime_per_library.{fmt}"
            if runtime_per_library_plot_file.is_file():
                assert force, f"File already exists: {runtime_per_library_plot_file}"
            runtime_per_library_plot.get_figure().savefig(
                runtime_per_library_plot_file,
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()
            # object
            object_runtime_df = agg_library_runtime(runtime_df, symbol_map_df, by="object", col="rel_weight")
            runtime_per_object_data = generate_pie_data(object_runtime_df, x="object", y="rel_weight", topk=topk)
            runtime_per_object_plot = plot_pie_data(
                runtime_per_object_data,
                "rel_weight",
                threshold=threshold,
                legend=legend,
                title="Runtime per Object",
            )
            runtime_per_object_plot_file = plots_dir / f"runtime_per_object.{fmt}"
            if runtime_per_object_plot_file.is_file():
                assert force, f"File already exists: {runtime_per_object_plot_file}"
            runtime_per_object_plot.get_figure().savefig(
                runtime_per_object_plot_file,
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()
    if instrs_hist_df is not None:
        instrs_hist_data = generate_pie_data(instrs_hist_df, x="instr", y="rel_count", topk=topk)
        instrs_hist_plot = plot_pie_data(
            instrs_hist_data,
            "rel_count",
            threshold=threshold,
            legend=legend,
            title="Runtime per Instr",
        )
        instrs_hist_plot_file = plots_dir / f"runtime_per_instr.{fmt}"
        if instrs_hist_plot_file.is_file():
            assert force, f"File already exists: {instrs_hist_plot_file}"
        instrs_hist_plot.get_figure().savefig(
            instrs_hist_plot_file,
            bbox_inches="tight",
            dpi=300,
        )
        plt.close()
    if opcodes_hist_df is not None:
        opcodes_hist_data = generate_pie_data(opcodes_hist_df, x="opcode", y="rel_count", topk=topk)
        opcodes_hist_plot = plot_pie_data(
            opcodes_hist_data,
            "rel_count",
            threshold=threshold,
            legend=legend,
            title="Runtime per Opcode",
        )
        opcodes_hist_plot_file = plots_dir / f"runtime_per_opcode.{fmt}"
        if opcodes_hist_plot_file.is_file():
            assert force, f"File already exists: {opcodes_hist_plot_file}"
        opcodes_hist_plot.get_figure().savefig(
            opcodes_hist_plot_file,
            bbox_inches="tight",
            dpi=300,
        )
        plt.close()

    # attrs = {
    #     # "elf_file": elf_artifact.name,
    #     "kind": "plot",
    #     "by": __name__,
    # }

    # artifact = TableArtifact("choices", choices_df, attrs=attrs)
    # sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    create_runtime_pie_plots(
        sess,
        threshold=args.threshold,
        topk=args.topk,
        legend=args.legend,
        force=args.force,
        fmt=args.fmt,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--legend", action="store_true")
    parser.add_argument("--threshold", type=float, default=0.1)
    parser.add_argument("--fmt", type=str, choices=["jpg", "png", "pdf"], default="jpg")
    parser.add_argument("--topk", type=int, default=9)
    # TODO: !
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# import time
import io
import sys
import itertools
import pandas as pd
import argparse
from pathlib import Path
from typing import Optional

from tqdm import tqdm

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import InstrTraceArtifact, TraceArtifact
from isaac_toolkit.logging import get_logger, set_log_level
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

logger = get_logger()


# TODO: logger


def process_df(df, operands: bool = False):
    df = df[df[0].str.contains(r" \(0x")]
    df = df[~df[0].str.contains("exception")]
    df = df[~df[0].str.contains("tval")]
    df[["core", "rest"]] = df[0].str.split(":", n=1, expand=True)
    df.drop(columns=[0], inplace=True)
    df["core"] = df["core"].apply(lambda x: int(x.split(" ", 1)[-1].strip()))
    df[["pc", "bytecode", "instr", "args"]] = df["rest"].apply(lambda x: x.strip()).str.split(" ", n=3, expand=True)
    df.drop(columns=["rest"], inplace=True)
    if operands:
        raise NotImplementedError
    else:
        df.drop(columns=["args"], inplace=True)
    df["bytecode"] = df["bytecode"].apply(lambda x: x[1:-1])
    df["bytecode"] = df["bytecode"].apply(lambda x: int(x, 0))
    df["pc"] = df["pc"].apply(lambda x: int(x, 0))

    df["pc"] = pd.to_numeric(df["pc"])
    df["instr"] = df["instr"].astype("category")

    def detect_size(instr):
        # if bytecode[:2] == "0x":
        #     return len(bytecode[2:]) // 2
        # elif bytecode[:2] == "0b":
        #     return len(bytecode[2:]) // 8
        # else:
        #     assert len(set(bytecode)) == 2
        #     return len(bytecode) // 8

        # Spike does not use shorter byte codes for compressed instrs
        # Alternatives: check LSBs or instr prefix
        if instr.startswith("c."):
            return 2
        return 4

    df["size"] = df["instr"].apply(detect_size)
    df["size"] = df["size"].astype("category")

    df["bytecode"] = pd.to_numeric(df["bytecode"])
    return df


def parse_and_process(chunk_bytes, operands: bool = False):
    # df = pd.read_csv(input_file, sep="@", header=None, chunksize=2**22) as reader:
    df = pd.read_csv(io.BytesIO(chunk_bytes), header=None, sep="@")
    return process_df(df, operands=operands)


def chunk_iter(path, chunk_size=2**22):
    f = open(path, "rb")

    with f:
        buf = b""
        while True:
            data = f.read(chunk_size)
            if not data:
                break
            buf += data
            last_nl = buf.rfind(b"\n")
            if last_nl == -1:
                continue
            yield buf[: last_nl + 1]
            buf = buf[last_nl + 1 :]
        if buf:
            yield buf


def load_instr_trace(
    sess: Session,
    input_file: Path,
    force: bool = False,
    operands: bool = False,
    progress: bool = False,
    num_workers: Optional[int] = None,
    executor: str = "process_pool",
    chunk_size: int = 2**22,
):
    logger.info("Loading Spike instruction trace...")
    assert input_file.is_file(), f"File not found: {input_file}"
    name = input_file.name
    # df = pd.read_csv(input_file, sep=":", names=["pc", "rest"])
    dfs = []
    if True:
        executor_map = {
            "thread_pool": ThreadPoolExecutor,
            "process_pool": ProcessPoolExecutor,
        }
        executor_cls = executor_map.get(executor)
        assert executor_cls is not None, f"Unsupported Executor: {executor}"
        with executor_cls(max_workers=num_workers) as executor:  # tune workers
            dfs_ = list(
                tqdm(
                    executor.map(
                        parse_and_process,
                        chunk_iter(input_file, chunk_size=chunk_size),
                        itertools.repeat(operands),
                    ),
                    disable=not progress,
                )
            )
            dfs += dfs_
    else:
        with pd.read_csv(input_file, sep="@", header=None, chunksize=2**22) as reader:
            for df in tqdm(reader, disable=False):
                df = df[df[0].str.contains(r" \(0x")]
                df[["core", "rest"]] = df[0].str.split(":", n=1, expand=True)
                df.drop(columns=[0], inplace=True)
                df["core"] = df["core"].apply(lambda x: int(x.split(" ", 1)[-1].strip()))
                df[["pc", "bytecode", "instr", "args"]] = (
                    df["rest"].apply(lambda x: x.strip()).str.split(" ", n=3, expand=True)
                )
                df.drop(columns=["rest"], inplace=True)
                if operands:
                    raise NotImplementedError
                else:
                    df.drop(columns=["args"], inplace=True)
                df["bytecode"] = df["bytecode"].apply(lambda x: x[1:-1])
                df["bytecode"] = df["bytecode"].apply(lambda x: int(x, 0))
                df["pc"] = df["pc"].apply(lambda x: int(x, 0))

                df["pc"] = pd.to_numeric(df["pc"])
                df["instr"] = df["instr"].astype("category")

                def detect_size(instr):
                    # if bytecode[:2] == "0x":
                    #     return len(bytecode[2:]) // 2
                    # elif bytecode[:2] == "0b":
                    #     return len(bytecode[2:]) // 8
                    # else:
                    #     assert len(set(bytecode)) == 2
                    #     return len(bytecode) // 8

                    # Spike does not use shorter byte codes for compressed instrs
                    # Alternatives: check LSBs or instr prefix
                    if instr.startswith("c."):
                        return 2
                    return 4

                df["size"] = df["instr"].apply(detect_size)
                df["size"] = df["size"].astype("category")

                df["bytecode"] = pd.to_numeric(df["bytecode"])

                dfs.append(df)

    df = pd.concat(dfs, axis=0)
    df["instr"] = df["instr"].astype("category")
    df["size"] = df["size"].astype("category")
    df["core"] = df["core"].astype("category")
    df["pc"] = pd.to_numeric(df["pc"], downcast="unsigned")
    df["is_trap"] = df["pc"] >= 0x80000000  # TODO: check if assumption can be made. PK only?
    df["bytecode"] = df["bytecode"].astype("category")
    df.reset_index(drop=True, inplace=True)

    attrs = {
        "simulator": "spike",
        "cpu_arch": "unknown",
        "by": "isaac_toolkit.frontend.instr_trace.spike",
    }
    if operands:
        operands_trace_df = df[["instr", "operands"]]
        df.drop(columns=["operands"], inplace=True)
        operands_artifact = TraceArtifact("operands_trace", operands_trace_df, attrs=attrs)
        sess.add_artifact(operands_artifact, override=force)
    artifact = InstrTraceArtifact(name, df, attrs=attrs)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    input_file = Path(args.file)
    load_instr_trace(
        sess,
        input_file,
        force=args.force,
        operands=args.operands,
        progress=args.progress,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--operands", action="store_true")
    parser.add_argument("--progress", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List, Optional, Union


import sys
import logging
import argparse
from pathlib import Path

import yaml
import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import filter_artifacts


logging.basicConfig(level=logging.DEBUG)  # TODO
logger = logging.getLogger(__name__)


def generate_ranges_yaml(
    sess: Session,
    output: Optional[Union[str, Path]] = None,
    force: bool = False,
    filter_funcs: Optional[List[str]] = None,
    sort_by: Optional[str] = None,
    sort_ascending: bool = False,
    topk: Optional[int] = None,
):
    artifacts = sess.artifacts

    bb_trace_artifacts = filter_artifacts(artifacts, lambda x: x.name == "bb_trace")
    assert len(bb_trace_artifacts) == 1
    bb_trace_artifact = bb_trace_artifacts[0]
    bb_trace_df = bb_trace_artifact.df

    unique_bbs_artifacts = filter_artifacts(artifacts, lambda x: x.name == "unique_bbs")
    assert len(unique_bbs_artifacts) == 1
    unique_bbs_artifact = unique_bbs_artifacts[0]
    unique_bbs_df = unique_bbs_artifact.df.copy()
    print("unique_bbs_df", unique_bbs_df)
    # unique_bbs_df

    if filter_funcs is not None and len(filter_funcs) > 0:
        unique_bbs_df = unique_bbs_df[unique_bbs_df["func"].isin(filter_funcs)]

    unique_bbs_df["func_bb_idx"] = unique_bbs_df.groupby("func").cumcount()
    unique_bbs_df["runtime_weight"] = unique_bbs_df["freq"] * unique_bbs_df["num_instrs"]

    if sort_by is not None:
        unique_bbs_df.sort_values(sort_by, inplace=True, ascending=sort_ascending)
    if topk is not None:
        unique_bbs_df = unique_bbs_df.iloc[:topk]
    print("unique_bbs_df", unique_bbs_df)

    print("bb_trace_df", bb_trace_df)
    # TODO: expose strategy for selecting which call (first, last, first+last, random, single, multiple,...)
    # per_bb_trace_df = bb_trace_df.groupby("bb_idx").first()
    per_bb_trace_df = bb_trace_df.groupby("bb_idx").last()
    print("per_bb_trace_df", per_bb_trace_df)

    merged_df = pd.merge(per_bb_trace_df, unique_bbs_df, how="inner", left_index=True, right_index=True)
    KEEP_COLS = ["bb_call", "trace_idx", "num_instrs", "size", "func", "func_bb_idx"]
    if sort_by is not None:
        KEEP_COLS.append(sort_by)
        merged_df.sort_values(sort_by, inplace=True, ascending=sort_ascending)
    merged_df = merged_df[KEEP_COLS]
    print("merged_df", merged_df)

    ranges_data = []

    for bb_idx, bb_row in merged_df.iterrows():
        func = bb_row.func
        start = bb_row.trace_idx
        num_instrs = bb_row.num_instrs
        func_bb_idx = bb_row.func_bb_idx
        call = bb_row.bb_call
        end = start + num_instrs - 1
        # name = f"{func}@bb{func_bb_idx}-BB{bb_idx}-I{call}"
        # TODO: add func_bb_idx to artifact!
        name = f"bb{func_bb_idx}@{func}-BB{bb_idx}-I{call}"
        new = [name, start, end]
        ranges_data.append(new)

    yaml_data = {"ranges": ranges_data}

    if output is None:
        profile_dir = sess.directory / "output"
        profile_dir.mkdir(exist_ok=True)
        out_name = "ranges.yml"
        output = profile_dir / out_name
    if Path(output).is_file():
        assert force, f"Output file '{output}' already exists. Use --force to override!"
    with open(output, "w") as f:
        yaml.dump(yaml_data, f, default_flow_style=False)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    generate_ranges_yaml(
        sess,
        output=args.output,
        force=args.force,
        filter_funcs=args.func_filter.split(",") if args.func_filter is not None else None,
        sort_by=args.sort_by,
        sort_ascending=args.ascending,
        topk=args.topk,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--output", default=None)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--func-filter", type=str, default=None)  # comma-separated
    parser.add_argument("--sort-by", type=str, default=None)
    parser.add_argument("--topk", type=int, default=None)
    parser.add_argument("--ascending", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import argparse
from pathlib import Path

import pandas as pd


def analyze_compare(report, mem_report=None, output=None, print_df=False):
    COLS = ["Model", "Arch", "Run Instructions", "Run Instructions (rel.)"]
    MEM_COLS = ["Model", "Arch", "Total ROM", "Total RAM", "ROM code", "ROM code (rel.)"]
    COMMON_COLS = list(set(COLS) & set(MEM_COLS))

    report_file = Path(report)
    assert report_file.is_file()
    report_df = pd.read_csv(report_file)[COLS]
    # print(report_df, report_df.columns)

    if mem_report:
        mem_report_file = Path(mem_report)
        assert mem_report_file.is_file()
        mem_report_df = pd.read_csv(mem_report_file)[MEM_COLS]
        # print(mem_report_df, mem_report_df.columns)
        report_df = report_df.merge(mem_report_df, on=COMMON_COLS)

    # print(report_df, report_df.columns)

    df = report_df

    assert print_df or output is not None

    if print_df:
        with pd.option_context("display.max_columns", None):
            print(df)

    if output is not None:
        # TODO: assert csv suffix?
        df.to_csv(output)


def handle(args):
    analyze_compare(
        args.report,
        mem_report=args.mem_report,
        output=args.output,
        print_df=args.print_df,
    )


def get_parser():
    parser = argparse.ArgumentParser(description="Collect metrics from MLonMCU Report")
    parser.add_argument("report", help="Report CSV file")
    parser.add_argument("--mem-report", default=None, help="Memory report CSV file")
    # parser.add_argument("--mem", action="store_true", help="Compare mem instead of Runtime")
    # parser.add_argument("-i", "--isax", default="auto", help="ISAX name")
    parser.add_argument("-o", "--output", default=None, help="Output file")
    parser.add_argument("--print-df", action="store_true", help="Print DataFrame")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

from . import Session


def get_summary(sess):
    config_text = sess.config.to_yaml()
    # inputs_text = "\n".join(f" - {artifact.summary()}" for artifact in sess.inputs)
    # outputs_text = "\n".join(f" - {artifact.summary()}" for artifact in sess.outputs)
    # temps_text = "\n".join(f" - {artifact.summary()}" for artifact in sess.temps)
    artifacts_text = "\n".join(f" - {artifact.summary()}" for artifact in sess.artifacts)
    return f"""Summary of ISAAC session {sess.directory}

Config:
```
---
{config_text}
```

Artifacts:
{artifacts_text}
"""


def handle_summary(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    text = get_summary(sess)
    assert len(text) > 0
    print(text)


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle_summary(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
riscv_branch_instrs = [
    "j",  # pseudo
    "jr",  # pseudo
    "call",  # pseudo
    "tail",  # pseudo
    "jal",
    "beq",
    "beqz",  # pseudo
    "bne",
    "blt",
    "bltz",  # pseudo
    "bgt",  # pseudo
    "bgtz",  # pseudo
    "bge",  # pseudo
    "bgez",  # pseudo
    "ble",
    "blez",  # new?
    "bltu",
    "bgtu",  # pseudo
    "bgtu",  # pseudo
    "bgeu",  # pseudo
    "bleu",
    "ecall",
    "bnez",  # bseudo
    "cbnez",
    "c.bnez",
    "cj",  # pseudo
    "cbeqz",
    "cjal",
    "c.j",
    "c.j",
    "c.beqz",
    "c.jal",
    # "ret",  # pseudo
    # "mret",  # pseudo
    # "sret",  # pseudo
    # "uret",  # pseudo
    # "jalr"  # return if rd=x0, rs1=x1
    # "c.jr",  # return if rs1=x1
    # "c.jalr",  # return if  rs1=x1
]
riscv_return_instrs = ["jalr", "cjalr", "cjr", "c.jr", "c.jalr", "ret", "mret", "sret", "uret"]  # TODO
